#!/usr/bin/env python
'''
mavproxy - a MAVLink proxy program

Copyright Andrew Tridgell 2011
Released under the GNU GPL version 3 or later

'''

import sys, os, time, socket, signal
import fnmatch, errno, threading
import serial
import traceback
import select
import shlex
import math
import platform
import json
import struct
import glob

try:
    reload
except NameError:
    try:
        from importlib import reload
    except ImportError:
        from imp import reload

try:
    import queue as Queue
except ImportError:
    import Queue

from builtins import input

from MAVProxy.modules.lib import textconsole
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import rline
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import dumpstacks
from MAVProxy.modules.lib import mp_substitute
from MAVProxy.modules.lib import multiproc
from MAVProxy.modules.mavproxy_link import preferred_ports

# adding all this allows pyinstaller to build a working windows executable
# note that using --hidden-import does not work for these modules
try:
      multiproc.freeze_support()
      from pymavlink import mavwp, mavutil
      import matplotlib, HTMLParser
except Exception:
      pass

# screensaver dbus syntax swiped from
# https://stackoverflow.com/questions/10885337/inhibit-screensaver-with-python
screensaver_interface = None
screensaver_cookie = None
try:
    import atexit
    import dbus
    bus = dbus.SessionBus()
    saver = bus.get_object('org.freedesktop.ScreenSaver', '/ScreenSaver')
    screensaver_interface = dbus.Interface(saver, dbus_interface='org.freedesktop.ScreenSaver')
    if screensaver_cookie is not None:
        atexit.register(saver_interface.UnInhibit, [screensaver_cookie])
except Exception as e:
    pass

if __name__ == '__main__':
      multiproc.freeze_support()

#The MAVLink version being used (None, "1.0", "2.0")
mavversion = None

class MPStatus(object):
    '''hold status information about the mavproxy'''
    def __init__(self):
        self.gps	 = None
        self.msgs = {}
        self.msg_count = {}
        self.counters = {'MasterIn' : [], 'MasterOut' : 0, 'FGearIn' : 0, 'FGearOut' : 0, 'Slave' : 0}
        self.bytecounters = {'MasterIn': []}
        self.setup_mode = opts.setup
        self.mav_error = 0
        self.altitude = 0
        self.last_distance_announce = 0.0
        self.exit = False
        self.flightmode = 'MAV'
        self.last_mode_announce = 0
        self.last_mode_announced = 'MAV'
        self.logdir = None
        self.last_heartbeat = 0
        self.last_message = 0
        self.heartbeat_error = False
        self.last_apm_msg = None
        self.last_apm_msg_time = 0
        self.statustexts_by_sysidcompid = {}
        self.highest_msec = {}
        self.have_gps_lock = False
        self.lost_gps_lock = False
        self.last_gps_lock = 0
        self.watch = None
        self.last_streamrate1 = -1
        self.last_streamrate2 = -1
        self.last_seq = 0
        self.armed = False
        self.last_bytecounter_calc = 0

    class ByteCounter(object):
        def __init__(self):
            self.total_count = 0
            self.current_count = 0
            self.buckets = []
            self.max_buckets = 10  # 10 seconds

        def update(self, bytecount):
            self.total_count += bytecount
            self.current_count += bytecount

        def rotate(self):
            '''move current count into a bucket, zero count'''
            # huge assumption made that we're called rapidly enough to
            # not need to rotate multiple buckets.
            self.buckets.append(self.current_count)
            self.current_count = 0
            if len(self.buckets) > self.max_buckets:
                self.buckets = self.buckets[-self.max_buckets:]

        def rate(self):
            if len(self.buckets) == 0:
                return 0
            total = 0
            for bucket in self.buckets:
                total += bucket
            return total/float(len(self.buckets))

        def total(self):
            return self.total_count

    def update_bytecounters(self):
        '''rotate bytecounter buckets if required'''
        now = time.time()
        time_delta = now - self.last_bytecounter_calc
        if time_delta < 1:
            return
        self.last_bytecounter_calc = now

        for counter in self.bytecounters['MasterIn']:
            counter.rotate()

    def show(self, f, pattern=None, verbose=False):
        '''write status to status.txt'''
        if pattern is None:
            f.write('Counters: ')
            for c in self.counters:
                f.write('%s:%s ' % (c, self.counters[c]))
            f.write('\n')
            f.write('MAV Errors: %u\n' % self.mav_error)
            f.write(str(self.gps)+'\n')
        for m in sorted(self.msgs.keys()):
            if pattern is not None:
                if not fnmatch.fnmatch(str(m).upper(), pattern.upper()):
                    continue
                if getattr(self.msgs[m], '_instance_field', None) is not None and m.find('[') == -1 and pattern.find('*') != -1:
                    # only show instance versions for patterns
                    continue
            msg = None
            sysid = mpstate.settings.target_system
            for mav in mpstate.mav_master:
                if not sysid in mav.sysid_state:
                    continue
                if not m in mav.sysid_state[sysid].messages:
                    continue
                msg2 = mav.sysid_state[sysid].messages[m]
                if msg is None or msg2._timestamp > msg._timestamp:
                    msg = msg2
            if msg is None:
                continue
            if verbose:
                try:
                    mavutil.dump_message_verbose(f, msg)
                    f.write("\n")
                except AttributeError as e:
                    if "has no attribute 'dump_message_verbose'" in str(e):
                        print("pymavlink update required for --verbose")
                    else:
                        raise e
            else:
                f.write("%u: %s\n" % (self.msg_count[m], str(msg)))

    def write(self):
        '''write status to status.txt'''
        f = open('status.txt', mode='w')
        self.show(f)
        f.close()

def say_text(text, priority='important'):
    '''text output - default function for say()'''
    mpstate.console.writeln(text)

def say(text, priority='important'):
    '''text and/or speech output'''
    mpstate.functions.say(text, priority)

def add_input(cmd, immediate=False):
    '''add some command input to be processed'''
    if immediate:
        process_stdin(cmd)
    else:
        mpstate.input_queue.put(cmd)

class MAVFunctions(object):
    '''core functions available in modules'''
    def __init__(self):
        self.process_stdin = add_input
        self.param_set = param_set
        self.get_mav_param = get_mav_param
        self.say = say_text
        # input handler can be overridden by a module
        self.input_handler = None

class MPState(object):
    '''holds state of mavproxy'''
    def __init__(self):
        self.console = textconsole.SimpleConsole()
        self.map = None
        self.map_functions = {}
        self.click_location = None
        self.click_time = None
        self.vehicle_type = None
        self.vehicle_name = None
        from MAVProxy.modules.lib.mp_settings import MPSettings, MPSetting
        self.settings = MPSettings(
            [ MPSetting('link', int, 1, 'Primary Link', tab='Link', range=(0,100), increment=1),
              MPSetting('streamrate', int, 4, 'Stream rate link1', range=(-1,500), increment=1),
              MPSetting('streamrate2', int, 4, 'Stream rate link2', range=(-1,500), increment=1),
              MPSetting('heartbeat', float, 1, 'Heartbeat rate (Hz)', range=(0,100), increment=0.1),
              MPSetting('mavfwd', bool, True, 'Allow forwarded control'),
              MPSetting('mavfwd_disarmed', bool, True, 'Allow forwarded control when disarmed'),
              MPSetting('mavfwd_rate', bool, False, 'Allow forwarded rate control'),
              MPSetting('mavfwd_link', int, -1, 'Forward to a specific link'),
              MPSetting('shownoise', bool, True, 'Show non-MAVLink data'),
              MPSetting('baudrate', int, opts.baudrate, 'baudrate for new links', range=(0,10000000), increment=1),
              MPSetting('rtscts', bool, opts.rtscts, 'enable flow control'),
              MPSetting('select_timeout', float, 0.01, 'select timeout'),

              MPSetting('altreadout', int, 10, 'Altitude Readout',
                        range=(0,100), increment=1, tab='Announcements'),
              MPSetting('distreadout', int, 200, 'Distance Readout', range=(0,10000), increment=1),

              MPSetting('moddebug', int, opts.moddebug, 'Module Debug Level', range=(0,3), increment=1, tab='Debug'),
              MPSetting('script_fatal', bool, False, 'fatal error on bad script', tab='Debug'),
              MPSetting('compdebug', int, 0, 'Computation Debug Mask', range=(0,3), tab='Debug'),
              MPSetting('flushlogs', bool, False, 'Flush logs on every packet'),
              MPSetting('requireexit', bool, False, 'Require exit command'),
              MPSetting('wpupdates', bool, True, 'Announce waypoint updates'),
              MPSetting('wpterrainadjust', bool, True, 'Adjust alt of moved wp using terrain'),
              MPSetting('wp_use_mission_int', bool, True, 'use MISSION_ITEM_INT messages'),

              MPSetting('basealt', int, 0, 'Base Altitude', range=(0,30000), increment=1, tab='Altitude'),
              MPSetting('wpalt', int, 100, 'Default WP Altitude', range=(0,10000), increment=1),
              MPSetting('rallyalt', int, 90, 'Default Rally Altitude', range=(0,10000), increment=1),
              MPSetting('terrainalt', str, 'Auto', 'Use terrain altitudes', choice=['Auto','True','False']),
              MPSetting('guidedalt', int, 100, 'Default "Fly To" Altitude', range=(0,10000), increment=1),
              MPSetting('rally_breakalt', int, 40, 'Default Rally Break Altitude', range=(0,10000), increment=1),
              MPSetting('rally_flags', int, 0, 'Default Rally Flags', range=(0,10000), increment=1),

              MPSetting('source_system', int, 255, 'MAVLink Source system', range=(0,255), increment=1, tab='MAVLink'),
              MPSetting('source_component', int, 230, 'MAVLink Source component', range=(0,255), increment=1),
              MPSetting('target_system', int, 0, 'MAVLink target system', range=(0,255), increment=1),
              MPSetting('target_component', int, 0, 'MAVLink target component', range=(0,255), increment=1),
              MPSetting('state_basedir', str, None, 'base directory for logs and aircraft directories'),
              MPSetting('allow_unsigned', bool, True, 'whether unsigned packets will be accepted'),

              MPSetting('dist_unit', str, 'm', 'distance unit', choice=['m', 'nm', 'miles'], tab='Units'),
              MPSetting('height_unit', str, 'm', 'height unit', choice=['m', 'feet']),
              MPSetting('speed_unit', str, 'm/s', 'height unit', choice=['m/s', 'knots', 'mph']),

              MPSetting('fwdpos', bool, False, 'Forward GLOBAL_POSITION_INT on all links'),
              MPSetting('checkdelay', bool, True, 'check for link delay'),
              MPSetting('param_ftp', bool, True, 'try ftp for parameter download'),
              MPSetting('param_docs', bool, True, 'show help for parameters'),

              MPSetting('vehicle_name', str, '', 'Vehicle Name', tab='Vehicle'),

              MPSetting('sys_status_error_warn_interval', int, 30, 'interval to warn of autopilot software failure'),

              MPSetting('inhibit_screensaver_when_armed', bool, False, 'inhibit screensaver while vehicle armed'),

              MPSetting('timeout', int, 5, 'Number of seconds with no packets for a link to considered down', range=(0,255), increment=1),

            ])

        self.completions = {
            "script"         : ["(FILENAME)"],
            "set"            : ["(SETTING)"],
            "status"         : ["(VARIABLE)"],
            "module"    : ["list",
                           "load (AVAILMODULES)",
                           "<unload|reload> (LOADEDMODULES)"]
            }

        self.status = MPStatus()

        # master mavlink device
        self.mav_master = None

        # mavlink outputs
        self.mav_outputs = []
        self.sysid_outputs = {}

        # Mapping of all detected sysid's to links
        # Key is link id, value is all detected sysid's/compid's in that link
        # Dict of self.vehicle_link_map[linknumber] = set([(sysid1,compid1), (sysid2,compid2), ...])
        self.vehicle_link_map = {}

        # SITL output
        self.sitl_output = None

        self.mav_param_by_sysid = {}
        self.mav_param_by_sysid[(self.settings.target_system,self.settings.target_component)] = mavparm.MAVParmDict()
        self.modules = []
        self.public_modules = {}
        self.functions = MAVFunctions()
        self.select_extra = {}
        self.continue_mode = False
        self.aliases = {}
        import platform
        self.system = platform.system()
        self.multi_instance = {}
        self.instance_count = {}
        self.is_sitl = False
        self.start_time_s = time.time()
        self.attitude_time_s = 0

    @property
    def mav_param(self):
        '''map mav_param onto the current target system parameters'''
        compid = self.settings.target_component
        if compid == 0:
            compid = 1
        sysid = (self.settings.target_system, compid)
        if not sysid in self.mav_param_by_sysid:
            self.mav_param_by_sysid[sysid] = mavparm.MAVParmDict()
        return self.mav_param_by_sysid[sysid]

    def module(self, name):
        '''Find a public module (most modules are private)'''
        if name in self.public_modules:
            return self.public_modules[name]
        return None

    def master(self, target_sysid = -1):
        '''return the currently chosen mavlink master object'''
        if len(self.mav_master) == 0:
              return None
        if self.settings.link > len(self.mav_master):
            self.settings.link = 1

        if target_sysid != -1:
            # if we're looking for a specific system ID then try to find best
            # link for that
            best_link = None
            best_timestamp = 0
            for m in self.mav_master:
                try:
                    tstamp = m.sysid_state[target_sysid].messages['HEARTBEAT']._timestamp
                except Exception:
                    continue
                if tstamp > best_timestamp:
                    best_link = m
                    best_timestamp = tstamp
            if best_link is not None:
                return best_link

        # try to use one with no link error
        if not self.mav_master[self.settings.link-1].linkerror:
            return self.mav_master[self.settings.link-1]
        for m in self.mav_master:
            if not m.linkerror:
                return m
        return self.mav_master[self.settings.link-1]

    def foreach_mav(self, sysid, compid, closure):
        # Send mavlink message only on all links that contain vehicle (sysid, compid)
        # More efficient than just blasting all links, when sending targetted messages
        # Also useful for sending messages to non-selected vehicles
        for linkNumber, vehicleList in self.vehicle_link_map.items():
            if (sysid, compid) not in vehicleList:
                continue
            closure(self.mav_master[linkNumber].mav)

    def notify_click(self):
        notify_mods = ['map', 'misseditor']
        for modname in notify_mods:
            mod = self.module(modname)
            if mod is not None:
                mod.click_updated()

    def click(self, latlng):
        if latlng is None:
            self.click_location = None
            self.click_time = None
            self.notify_click()
            return

        (lat, lng) = latlng
        if lat is None:
            print("Bad Lat")
            return
        if lng is None:
            print("Bad lng")
            return
        self.click_location = (lat, lng)
        self.click_time = time.time()
        self.notify_click()

def get_mav_param(param, default=None):
    '''return a EEPROM parameter value'''
    return mpstate.mav_param.get(param, default)

def param_set(name, value, retries=3):
    '''set a parameter'''
    name = name.upper()
    return mpstate.mav_param.mavset(mpstate.master(), name, value, retries=retries)

def cmd_script(args):
    '''run a script'''
    if len(args) < 1:
        print("usage: script <filename>")
        return

    run_script(args[0])

def cmd_set(args):
    '''control mavproxy options'''
    mpstate.settings.command(args)

def cmd_status(args):
    '''show status'''
    verbose = False
    if "--verbose" in args:
        verbose = True
        args = list(filter(lambda x : x != "--verbose", args))
    if len(args) == 0:
        mpstate.status.show(sys.stdout, pattern=None, verbose=verbose)
    else:
        for pattern in args:
            mpstate.status.show(sys.stdout, pattern=pattern, verbose=verbose)

def cmd_setup(args):
    mpstate.status.setup_mode = True
    mpstate.rl.set_prompt("")


def cmd_reset(args):
    print("Resetting master")
    mpstate.master().reset()

def cmd_click(args):
    '''synthesise click at lat/lon; no arguments is "unclick"'''
    if len(args) == 0:
        mpstate.click(None)
        return
    if len(args) < 2:
        print("click LAT_EXPRESSION LNG_EXPRESSION")
        return
    lat = mavutil.evaluate_expression(args[0], mpstate.master().messages)
    lng = mavutil.evaluate_expression(args[1], mpstate.master().messages)
    mpstate.click((lat, lng))

def cmd_watch(args):
    '''watch a mavlink packet pattern'''
    if len(args) == 0:
        mpstate.status.watch = None
        return
    mpstate.status.watch = args
    print("Watching %s" % mpstate.status.watch)

def generate_kwargs(args):
    kwargs = {}
    module_components = args.split(":{", 1)
    module_name = module_components[0]
    if (len(module_components) == 2 and module_components[1].endswith("}")):
        # assume json
        try:
            module_args = "{"+module_components[1]
            kwargs = json.loads(module_args)
        except ValueError as e:
            print('Invalid JSON argument: {0} ({1})'.format(module_args,
                                                           repr(e)))
    return (module_name, kwargs)

def get_exception_stacktrace(e):
    if sys.version_info[0] >= 3:
        ret = "%s\n" % e
        ret += ''.join(traceback.format_exception(type(e),
                                                  e,
                                                  tb=e.__traceback__))
        return ret
    return traceback.format_exc(e)

def load_module(modname, quiet=False, **kwargs):
    '''load a module'''
    modpaths = ['MAVProxy.modules.mavproxy_%s' % modname, modname]
    for (m,pm) in mpstate.modules:
        if m.name == modname and not modname in mpstate.multi_instance:
            if not quiet:
                print("module %s already loaded" % modname)
            # don't report an error
            return True
    ex = None
    for modpath in modpaths:
        try:
            m = import_package(modpath)
            reload(m)
            module = m.init(mpstate, **kwargs)
            if isinstance(module, mp_module.MPModule):
                mpstate.modules.append((module, m))
                if not quiet:
                    if kwargs:
                        print("Loaded module %s with kwargs = %s" % (modname, kwargs))
                    else:
                        print("Loaded module %s" % (modname,))
                return True
            else:
                ex = "%s.init did not return a MPModule instance" % modname
                break
        except ImportError as msg:
            ex = msg
            if mpstate.settings.moddebug > 1:
                print(get_exception_stacktrace(ex))
    help_traceback = ""
    if mpstate.settings.moddebug < 3:
        help_traceback = " Use 'set moddebug 3' in the MAVProxy console to enable traceback"
    print("Failed to load module: %s.%s" % (ex, help_traceback))
    return False

def unload_module(modname):
    '''unload a module'''
    for (m,pm) in mpstate.modules:
        if m.name == modname:
            if hasattr(m, 'unload'):
                t = threading.Thread(target=lambda : m.unload(), name="unload %s" % modname)
                t.start()
                t.join(timeout=5)
                if t.is_alive():
                    print("unload on module %s did not complete" % m.name)
                    mpstate.modules.remove((m,pm))
                    return False
            mpstate.modules.remove((m,pm))
            if modname in mpstate.public_modules:
                del mpstate.public_modules[modname]
            print("Unloaded module %s" % modname)
            return True
    print("Unable to find module %s" % modname)
    return False

def cmd_module(args):
    '''module commands'''
    usage = "usage: module <list|load|reload|unload>"
    if len(args) < 1:
        print(usage)
        return
    if args[0] == "list":
        mods = []
        for (m,pm) in mpstate.modules:
            mods.append(m)
        mods = sorted(mods, key=lambda m : m.name)
        for m in mods:
            print("%s: %s" % (m.name, m.description))
    elif args[0] == "load":
        if len(args) < 2:
            print("usage: module load <name>")
            return
        (modname, kwargs) = generate_kwargs(args[1])
        try:
            load_module(modname, **kwargs)
        except TypeError as ex:
            print(ex)
            print("%s module does not support keyword arguments"% modname)
            return
    elif args[0] == "reload":
        if len(args) < 2:
            print("usage: module reload <name>")
            return
        (modname, kwargs) = generate_kwargs(args[1])
        pmodule = None
        for (m,pm) in mpstate.modules:
            if m.name == modname:
                pmodule = pm
        if pmodule is None:
            print("Module %s not loaded" % modname)
            return
        if unload_module(modname):
            import zipimport
            try:
                reload(pmodule)
            except ImportError:
                clear_zipimport_cache()
                reload(pmodule)
            try:
                if load_module(modname, quiet=True, **kwargs):
                    print("Reloaded module %s" % modname)
            except TypeError:
                print("%s module does not support keyword arguments" % modname)
    elif args[0] == "unload":
        if len(args) < 2:
            print("usage: module unload <name>")
            return
        modname = os.path.basename(args[1])
        unload_module(modname)
    else:
        print(usage)


def cmd_alias(args):
    '''alias commands'''
    usage = "usage: alias <add|remove|list>"
    if len(args) < 1 or args[0] == "list":
        if len(args) >= 2:
            wildcard = args[1].upper()
        else:
            wildcard = '*'
        for a in sorted(mpstate.aliases.keys()):
            if fnmatch.fnmatch(a.upper(), wildcard):
                print("%-15s : %s" % (a, mpstate.aliases[a]))
    elif args[0] == "add":
        if len(args) < 3:
            print(usage)
            return
        a = args[1]
        mpstate.aliases[a] = ' '.join(args[2:])
    elif args[0] == "remove":
        if len(args) != 2:
            print(usage)
            return
        a = args[1]
        if a in mpstate.aliases:
            mpstate.aliases.pop(a)
        else:
            print("no alias %s" % a)
    else:
        print(usage)
        return


def clear_zipimport_cache():
    """Clear out cached entries from _zip_directory_cache.
    See http://www.digi.com/wiki/developer/index.php/Error_messages"""
    import sys, zipimport
    syspath_backup = list(sys.path)
    zipimport._zip_directory_cache.clear()

    # load back items onto sys.path
    sys.path = syspath_backup
    # add this too: see https://mail.python.org/pipermail/python-list/2005-May/353229.html
    sys.path_importer_cache.clear()

# http://stackoverflow.com/questions/211100/pythons-import-doesnt-work-as-expected
# has info on why this is necessary.

def import_package(name):
    """Given a package name like 'foo.bar.quux', imports the package
    and returns the desired module."""
    import zipimport
    try:
        mod = __import__(name)
    except ImportError:
        clear_zipimport_cache()
        mod = __import__(name)

    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


command_map = {
    'script'  : (cmd_script,   'run a script of MAVProxy commands'),
    'setup'   : (cmd_setup,    'go into setup mode'),
    'reset'   : (cmd_reset,    'reopen the connection to the MAVLink master'),
    'click'   : (cmd_click,    'set click location'),
    'status'  : (cmd_status,   'show status'),
    'set'     : (cmd_set,      'mavproxy settings'),
    'watch'   : (cmd_watch,    'watch a MAVLink pattern'),
    'module'  : (cmd_module,   'module commands'),
    'alias'   : (cmd_alias,    'command aliases')
    }

def shlex_quotes(value):
    '''see http://stackoverflow.com/questions/6868382/python-shlex-split-ignore-single-quotes'''
    lex = shlex.shlex(value)
    lex.quotes = '"'
    lex.whitespace_split = True
    lex.commenters = ''
    return list(lex)

def process_stdin(line):
    '''handle commands from user'''
    if line is None:
        sys.exit(0)

    # allow for modules to override input handling
    if mpstate.functions.input_handler is not None:
          mpstate.functions.input_handler(line)
          return

    line = line.strip()

    if mpstate.status.setup_mode:
        # in setup mode we send strings straight to the master
        if line == '.':
            mpstate.status.setup_mode = False
            mpstate.status.flightmode = "MAV"
            mpstate.rl.set_prompt("MAV> ")
            return
        if line != '+++':
            line += '\r'
        for c in line:
            time.sleep(0.01)
            if sys.version_info.major >= 3:
                mpstate.master().write(bytes(c, "ascii"))
            else:
                mpstate.master().write(c)
        return

    if not line:
        return

    try:
        args = shlex_quotes(line)
    except Exception as e:
        print("Caught shlex exception: %s" % e.message);
        return

    cmd = args[0]
    while cmd in mpstate.aliases:
        line = mpstate.aliases[cmd]
        args = shlex.split(line) + args[1:]
        cmd = args[0]

    if cmd == 'help':
        k = command_map.keys()
        k = sorted(k)
        for cmd in k:
            (fn, help) = command_map[cmd]
            print("%-15s : %s" % (cmd, help))
        return
    if cmd == 'exit' and mpstate.settings.requireexit:
        mpstate.status.exit = True
        return

    if not cmd in command_map:
        for (m,pm) in mpstate.modules:
            if hasattr(m, 'unknown_command'):
                try:
                    if m.unknown_command(args):
                        return
                except Exception as e:
                    print("ERROR in command: %s" % str(e))
        print("Unknown command '%s'" % line)
        return
    (fn, help) = command_map[cmd]
    try:
        fn(args[1:])
    except Exception as e:
        print("ERROR in command %s: %s" % (args[1:], str(e)))
        if mpstate.settings.moddebug > 1:
            traceback.print_exc()


def process_master(m):
    '''process packets from the MAVLink master'''
    try:
        s = m.recv(16*1024)
    except Exception:
        time.sleep(0.1)
        return
    # prevent a dead serial port from causing the CPU to spin. The user hitting enter will
    # cause it to try and reconnect
    if len(s) == 0:
        time.sleep(0.1)
        return

    mpstate.status.bytecounters['MasterIn'][m.linknum].update(len(s))

    if (mpstate.settings.compdebug & 1) != 0:
        return

    if mpstate.logqueue_raw:
        mpstate.logqueue_raw.put(bytearray(s))

    if mpstate.status.setup_mode:
        if mpstate.system == 'Windows':
           # strip nsh ansi codes
           s = s.replace("\033[K","")
        if sys.version_info.major >= 3:
            sys.stdout.write(str(s, "ascii", "ignore"))
        else:
            sys.stdout.write(str(s))
        sys.stdout.flush()
        return
    
    global mavversion
    if m.first_byte and mavversion is None:
        m.auto_mavlink_version(s)
    msgs = m.mav.parse_buffer(s)
    if msgs:
        for msg in msgs:
            sysid = msg.get_srcSystem()
            if sysid in mpstate.sysid_outputs:
                  # the message has been handled by a specialised handler for this system
                  continue
            if getattr(m, '_timestamp', None) is None:
                m.post_message(msg)
            if msg.get_type() == "BAD_DATA":
                if opts.show_errors:
                    mpstate.console.writeln("MAV error: %s" % msg)
                mpstate.status.mav_error += 1



def process_mavlink(slave):
    '''process packets from MAVLink slaves, forwarding to the master'''
    try:
        buf = slave.recv()
    except socket.error:
        return
    try:
        global mavversion
        if slave.first_byte and mavversion is None:
            slave.auto_mavlink_version(buf)
        msgs = slave.mav.parse_buffer(buf)
    except mavutil.mavlink.MAVError as e:
        mpstate.console.error("Bad MAVLink slave message from %s: %s" % (slave.address, e.message))
        return
    if msgs is None:
        return
    allow_fwd = mpstate.settings.mavfwd
    if not allow_fwd and mpstate.settings.mavfwd_disarmed and not mpstate.master(-1).motors_armed():
        allow_fwd = True
    if mpstate.status.setup_mode:
        allow_fwd = False
    if allow_fwd:
        for m in msgs:
            target_sysid = getattr(m, 'target_system', -1)
            mbuf = m.get_msgbuf()
            if mpstate.settings.mavfwd_link > 0 and mpstate.settings.mavfwd_link <= len(mpstate.mav_master):
                mpstate.mav_master[mpstate.settings.mavfwd_link-1].write(mbuf)
            else:
                # find best link by sysid
                mpstate.master(target_sysid).write(mbuf)
            if mpstate.logqueue:
                usec = int(time.time() * 1.0e6)
                mpstate.logqueue.put(bytearray(struct.pack('>Q', usec) + m.get_msgbuf()))
            if mpstate.status.watch:
                for msg_type in mpstate.status.watch:
                    if fnmatch.fnmatch(m.get_type().upper(), msg_type.upper()):
                        mpstate.console.writeln('> '+ str(m))
                        break
    mpstate.status.counters['Slave'] += 1


def mkdir_p(dir):
    '''like mkdir -p'''
    if not dir:
        return
    if dir.endswith("/"):
        mkdir_p(dir[:-1])
        return
    if os.path.isdir(dir):
        return
    mkdir_p(os.path.dirname(dir))
    os.mkdir(dir)

def log_writer():
    '''log writing thread'''
    while True:
        mpstate.logfile_raw.write(bytearray(mpstate.logqueue_raw.get()))
        timeout = time.time() + 10
        while not mpstate.logqueue_raw.empty() and time.time() < timeout:
            mpstate.logfile_raw.write(mpstate.logqueue_raw.get())
        while not mpstate.logqueue.empty() and time.time() < timeout:
            mpstate.logfile.write(mpstate.logqueue.get())
        if mpstate.settings.flushlogs or time.time() >= timeout:
            mpstate.logfile.flush()
            mpstate.logfile_raw.flush()

# If state_basedir is NOT set then paths for logs and aircraft
# directories are relative to mavproxy's cwd
def log_paths():
    '''Returns tuple (logdir, telemetry_log_filepath, raw_telemetry_log_filepath)'''
    if opts.aircraft is not None:
        dirname = ""
        if opts.mission is not None:
            print(opts.mission)
            dirname += "%s/logs/%s/Mission%s" % (opts.aircraft, time.strftime("%Y-%m-%d"), opts.mission)
        else:
            dirname += "%s/logs/%s" % (opts.aircraft, time.strftime("%Y-%m-%d"))
        # dirname is currently relative.  Possibly add state_basedir:
        if mpstate.settings.state_basedir is not None:
            dirname = os.path.join(mpstate.settings.state_basedir,dirname)
        mkdir_p(dirname)
        highest = None
        for i in range(1, 10000):
            fdir = os.path.join(dirname, 'flight%u' % i)
            if not os.path.exists(fdir):
                break
            highest = fdir
        if mpstate.continue_mode and highest is not None:
            fdir = highest
        elif os.path.exists(fdir):
            print("Flight logs full")
            sys.exit(1)
        logname = 'flight.tlog'
        logdir = fdir
    else:
        logname = os.path.basename(opts.logfile)
        dir_path = os.path.dirname(opts.logfile)
        if not os.path.isabs(dir_path) and mpstate.settings.state_basedir is not None:
            dir_path = os.path.join(mpstate.settings.state_basedir,dir_path)

        logdir = dir_path

    mkdir_p(logdir)
    return (logdir,
            os.path.join(logdir, logname),
            os.path.join(logdir, logname + '.raw'))


def open_telemetry_logs(logpath_telem, logpath_telem_raw):
    '''open log files'''
    if opts.append_log or opts.continue_mode:
        mode = 'ab'
    else:
        mode = 'wb'

    try:
        mpstate.logfile = open(logpath_telem, mode=mode)
        mpstate.logfile_raw = open(logpath_telem_raw, mode=mode)
        print("Log Directory: %s" % mpstate.status.logdir)
        print("Telemetry log: %s" % logpath_telem)

        #make sure there's enough free disk space for the logfile (>200Mb)
        #statvfs doesn't work in Windows
        if platform.system() != 'Windows':
            stat = os.statvfs(logpath_telem)
            if stat.f_bfree*stat.f_bsize < 209715200:
                print("ERROR: Not enough free disk space for logfile")
                mpstate.status.exit = True
                return

        # use a separate thread for writing to the logfile to prevent
        # delays during disk writes (important as delays can be long if camera
        # app is running)
        t = threading.Thread(target=log_writer, name='log_writer')
        t.daemon = True
        t.start()
    except Exception as e:
        print("ERROR: opening log file for writing: %s" % e)
        mpstate.status.exit = True
        return


def set_stream_rates():
    '''set mavlink stream rates'''
    if (not msg_period.trigger() and
        mpstate.status.last_streamrate1 == mpstate.settings.streamrate and
        mpstate.status.last_streamrate2 == mpstate.settings.streamrate2):
        return
    mpstate.status.last_streamrate1 = mpstate.settings.streamrate
    mpstate.status.last_streamrate2 = mpstate.settings.streamrate2
    for master in mpstate.mav_master:
        if master.linknum == 0:
            rate = mpstate.settings.streamrate
        else:
            rate = mpstate.settings.streamrate2
        if rate != -1 and mpstate.settings.streamrate != -1:
            # Send to all detected vehicles in this link
            for (sysid, compid) in mpstate.vehicle_link_map[master.linknum]:
                master.mav.request_data_stream_send(sysid, compid,
                                                    mavutil.mavlink.MAV_DATA_STREAM_ALL,
                                                    rate, 1)

def check_link_status():
    '''check status of master links'''
    tnow = time.time()
    if mpstate.status.last_message != 0 and tnow > mpstate.status.last_message + mpstate.settings.timeout:
        say("no link")
        mpstate.status.heartbeat_error = True
    for master in mpstate.mav_master:
        if not master.linkerror and (tnow > master.last_message + mpstate.settings.timeout or master.portdead):
            say("link %s down" % (mp_module.MPModule.link_label(master)))
            master.linkerror = True

def send_heartbeat(master):
    if master.mavlink10():
        master.mav.heartbeat_send(mavutil.mavlink.MAV_TYPE_GCS, mavutil.mavlink.MAV_AUTOPILOT_INVALID,
                                  0, 0, 0)
    else:
        MAV_GROUND = 5
        MAV_AUTOPILOT_NONE = 4
        master.mav.heartbeat_send(MAV_GROUND, MAV_AUTOPILOT_NONE)

def periodic_tasks():
    '''run periodic checks'''
    if mpstate.status.setup_mode:
        return

    if (mpstate.settings.compdebug & 2) != 0:
        return

    if mpstate.settings.heartbeat != 0:
        heartbeat_period.frequency = mpstate.settings.heartbeat

    if heartbeat_period.trigger() and mpstate.settings.heartbeat != 0:
        mpstate.status.counters['MasterOut'] += 1
        for master in mpstate.mav_master:
            send_heartbeat(master)

    if heartbeat_check_period.trigger():
        check_link_status()

    set_stream_rates()

    mpstate.status.update_bytecounters()

    # call optional module idle tasks. These are called at several hundred Hz
    for (m,pm) in mpstate.modules:
        if hasattr(m, 'idle_task'):
            try:
                m.idle_task()
            except Exception as msg:
                if mpstate.settings.moddebug == 1:
                    print(msg)
                elif mpstate.settings.moddebug > 1:
                    print(get_exception_stacktrace(msg))

        # also see if the module should be unloaded:
        if m.needs_unloading:
            unload_module(m.name)

def main_loop():
    '''main processing loop'''

    global screensaver_cookie

    if not mpstate.status.setup_mode and not opts.nowait:
        for master in mpstate.mav_master:
            if master.linknum != 0:
                break
            print("Waiting for heartbeat from %s" % master.address)
            send_heartbeat(master)
            master.wait_heartbeat(timeout=0.1)
        set_stream_rates()

    while True:
        if mpstate is None or mpstate.status.exit:
            return

        # enable or disable screensaver:
        if (mpstate.settings.inhibit_screensaver_when_armed and
            screensaver_interface is not None):
            if mpstate.status.armed and screensaver_cookie is None:
                # now we can inhibit the screensaver
                screensaver_cookie = screensaver_interface.Inhibit("MAVProxy",
                                                             "Vehicle is armed")
            elif not mpstate.status.armed and screensaver_cookie is not None:
                # we can also restore it
                screensaver_interface.UnInhibit(screensaver_cookie)
                screensaver_cookie = None

        while not mpstate.input_queue.empty():
            line = mpstate.input_queue.get()
            mpstate.input_count += 1
            cmds = line.split(';')
            if len(cmds) == 1 and cmds[0] == "":
                  mpstate.empty_input_count += 1
            for c in cmds:
                process_stdin(c)

        for master in mpstate.mav_master:
            if master.fd is None:
                if master.port.inWaiting() > 0:
                    process_master(master)

        periodic_tasks()

        rin = []
        for master in mpstate.mav_master:
            if master.fd is not None and not master.portdead:
                rin.append(master.fd)
        for m in mpstate.mav_outputs:
            rin.append(m.fd)
        for sysid in mpstate.sysid_outputs:
            m = mpstate.sysid_outputs[sysid]
            rin.append(m.fd)
        if rin == []:
            time.sleep(0.0001)
            continue

        for fd in mpstate.select_extra:
            rin.append(fd)
        try:
            (rin, win, xin) = select.select(rin, [], [], mpstate.settings.select_timeout)
        except select.error:
            continue

        if mpstate is None:
            return

        for fd in rin:
            if mpstate is None:
                  return
            for master in mpstate.mav_master:
                  if fd == master.fd:
                        process_master(master)
                        if mpstate is None:
                              return
                        continue
            for m in mpstate.mav_outputs:
                if fd == m.fd:
                    process_mavlink(m)
                    if mpstate is None:
                          return
                    continue

            for sysid in mpstate.sysid_outputs:
                m = mpstate.sysid_outputs[sysid]
                if fd == m.fd:
                    process_mavlink(m)
                    if mpstate is None:
                          return
                    continue

            # this allow modules to register their own file descriptors
            # for the main select loop
            if fd in mpstate.select_extra:
                try:
                    # call the registered read function
                    (fn, args) = mpstate.select_extra[fd]
                    fn(args)
                except Exception as msg:
                    if mpstate.settings.moddebug == 1:
                        print(msg)
                    # on an exception, remove it from the select list
                    mpstate.select_extra.pop(fd)



def input_loop():
    '''wait for user input'''
    while mpstate.status.exit != True:
        try:
            line = mpstate.rl.input()
            mpstate.input_queue.put(line)
        except (EOFError, IOError):
            mpstate.status.exit = True


def run_script(scriptfile):
    '''run a script file'''
    try:
        f = open(scriptfile, mode='r')
    except Exception:
        return
    mpstate.console.writeln("Running script %s" % scriptfile)
    sub = mp_substitute.MAVSubstitute()
    for line in f:
        line = line.strip()
        if line == "" or line.startswith('#'):
            continue
        try:
            line = sub.substitute(line, os.environ)
        except mp_substitute.MAVSubstituteError as ex:
            print("Bad variable: %s" % str(ex))
            if mpstate.settings.script_fatal:
                sys.exit(1)
            continue
        if line.startswith('@'):
            line = line[1:]
        else:
            mpstate.console.writeln("-> %s" % line)
        process_stdin(line)
    f.close()
    
def set_mav_version(mav10, mav20, autoProtocol, mavversionArg):
    '''Set the Mavlink version based on commandline options'''
#    if(mav10 == True or mav20 == True or autoProtocol == True):
#        print("Warning: Using deprecated --mav10, --mav20 or --auto-protocol options. Use --mavversion instead")

    #sanity check the options
    if (mav10 == True or mav20 == True) and autoProtocol == True:
        print("Error: Can't have [--mav10, --mav20] and --auto-protocol both True")
        sys.exit(1)
    if mav10 == True and mav20 == True:
        print("Error: Can't have --mav10 and --mav20 both True")
        sys.exit(1)
    if mavversionArg is not None and (mav10 == True or mav20 == True or autoProtocol == True):
        print("Error: Can't use --mavversion with legacy (--mav10, --mav20 or --auto-protocol) options")
        sys.exit(1)

    #and set the specific mavlink version (False = autodetect)
    global mavversion
    if mavversionArg == "1.0" or mav10 == True:
        os.environ['MAVLINK09'] = '1'
        mavversion = "1"
    else:
        os.environ['MAVLINK20'] = '1'
        mavversion = "2"

if __name__ == '__main__':
    from optparse import OptionParser
    parser = OptionParser("mavproxy.py [options]")

    parser.add_option("--master", dest="master", action='append',
                      metavar="DEVICE[,BAUD]", help="MAVLink master port and optional baud rate",
                      default=[])
    parser.add_option("", "--force-connected", dest="force_connected", help="Use master even if initial connection fails",
                      action='store_true', default=False)
    parser.add_option("--out", dest="output", action='append',
                      metavar="DEVICE[,BAUD]", help="MAVLink output port and optional baud rate",
                      default=[])
    parser.add_option("--baudrate", dest="baudrate", type='int',
                      help="default serial baud rate", default=57600)
    parser.add_option("--sitl", dest="sitl",  default=None, help="SITL output port")
    parser.add_option("--streamrate",dest="streamrate", default=4, type='int',
                      help="MAVLink stream rate")
    parser.add_option("--source-system", dest='SOURCE_SYSTEM', type='int',
                      default=255, help='MAVLink source system for this GCS')
    parser.add_option("--source-component", dest='SOURCE_COMPONENT', type='int',
                      default=230, help='MAVLink source component for this GCS')
    parser.add_option("--target-system", dest='TARGET_SYSTEM', type='int',
                      default=0, help='MAVLink target master system')
    parser.add_option("--target-component", dest='TARGET_COMPONENT', type='int',
                      default=0, help='MAVLink target master component')
    parser.add_option("--logfile", dest="logfile", help="MAVLink master logfile",
                      default='mav.tlog')
    parser.add_option("-a", "--append-log", dest="append_log", help="Append to log files",
                      action='store_true', default=False)
    parser.add_option("--quadcopter", dest="quadcopter", help="use quadcopter controls",
                      action='store_true', default=False)
    parser.add_option("--setup", dest="setup", help="start in setup mode",
                      action='store_true', default=False)
    parser.add_option("--nodtr", dest="nodtr", help="disable DTR drop on close",
                      action='store_true', default=False)
    parser.add_option("--show-errors", dest="show_errors", help="show MAVLink error packets",
                      action='store_true', default=False)
    parser.add_option("--speech", dest="speech", help="use text to speech",
                      action='store_true', default=False)
    parser.add_option("--aircraft", dest="aircraft", help="aircraft name", default=None)
    parser.add_option("--cmd", dest="cmd", help="initial commands", default=None, action='append')
    parser.add_option("--console", action='store_true', help="use GUI console")
    if platform.system() == 'Windows':
        parser.add_option("--no-console", action='store_true', help="don't use GUI console")
    parser.add_option("--map", action='store_true', help="load map module")
    parser.add_option(
        '--load-module',
        action='append',
        default=[],
        help='Load the specified module. Can be used multiple times, or with a comma separated list')
    parser.add_option("--mav10", action='store_true', default=False, help="Use MAVLink protocol 1.0")
    parser.add_option("--mav20", action='store_true', default=False, help="Use MAVLink protocol 2.0")
    parser.add_option("--auto-protocol", action='store_true', default=False, help="Auto detect MAVLink protocol version")
    parser.add_option("--mavversion", type='choice', choices=['1.0', '2.0'] , help="Force MAVLink Version (1.0, 2.0). Otherwise autodetect version")
    parser.add_option("--nowait", action='store_true', default=False, help="don't wait for HEARTBEAT on startup")
    parser.add_option("-c", "--continue", dest='continue_mode', action='store_true', default=False, help="continue logs")
    parser.add_option("--dialect",  default="ardupilotmega", help="MAVLink dialect")
    parser.add_option("--rtscts",  action='store_true', help="enable hardware RTS/CTS flow control")
    parser.add_option("--moddebug",  type=int, help="module debug level", default=0)
    parser.add_option("--mission", dest="mission", help="mission name", default=None)
    parser.add_option("--daemon", action='store_true', help="run in daemon mode, do not start interactive shell")
    parser.add_option("--non-interactive", action='store_true', help="do not start interactive shell")
    parser.add_option("--profile", action='store_true', help="run the Yappi python profiler")
    parser.add_option("--state-basedir", default=None, help="base directory for logs and aircraft directories")
    parser.add_option("--version", action='store_true', help="version information")
    parser.add_option("--default-modules", default="log,signing,wp,rally,fence,ftp,param,relay,tuneopt,arm,mode,calibration,rc,auxopt,misc,cmdlong,battery,terrain,output,adsb,layout", help='default module list')

    (opts, args) = parser.parse_args()
    if len(args) != 0:
          print("ERROR: mavproxy takes no position arguments; got (%s)" % str(args))
          sys.exit(1)

    # warn people about ModemManager which interferes badly with APM and Pixhawk
    if os.path.exists("/usr/sbin/ModemManager"):
        print("WARNING: You should uninstall ModemManager as it conflicts with APM and Pixhawk")

    #set the Mavlink version, if required
    set_mav_version(opts.mav10, opts.mav20, opts.auto_protocol, opts.mavversion)

    from pymavlink import mavutil, mavparm
    mavutil.set_dialect(opts.dialect)

    #version information
    if opts.version:
        #pkg_resources doesn't work in the windows exe build, so read the version file
        try:
            import pkg_resources
            version = pkg_resources.require("mavproxy")[0].version
        except:
            start_script = mp_util.dot_mavproxy("version.txt")
            f = open(start_script, 'r')
            version = f.readline()

        print("MAVProxy is a modular ground station using the mavlink protocol")
        print("MAVProxy Version: " + version)
        sys.exit(0)

    # global mavproxy state
    mpstate = MPState()
    mpstate.status.exit = False
    mpstate.command_map = command_map
    mpstate.continue_mode = opts.continue_mode
    # queues for logging
    mpstate.logqueue = Queue.Queue()
    mpstate.logqueue_raw = Queue.Queue()


    if opts.speech:
        # start the speech-dispatcher early, so it doesn't inherit any ports from
        # modules/mavutil
        load_module('speech')

    serial_list = mavutil.auto_detect_serial(preferred_list=preferred_ports)
    serial_list.sort(key=lambda x: x.device)

    # remove OTG2 ports for dual CDC
    if len(serial_list) == 2 and serial_list[0].device.startswith("/dev/serial/by-id"):
        if serial_list[0].device[:-1] == serial_list[1].device[0:-1]:
            serial_list.pop(1)

    if not opts.master:
        print('Auto-detected serial ports are:')
        for port in serial_list:
              print("%s" % port)

    # container for status information
    mpstate.settings.target_system = opts.TARGET_SYSTEM
    mpstate.settings.target_component = opts.TARGET_COMPONENT

    mpstate.mav_master = []

    mpstate.rl = rline.rline("MAV> ", mpstate)

    def quit_handler(signum = None, frame = None):
        #print('Signal handler called with signal', signum)
        if mpstate.status.exit:
            print('Clean shutdown impossible, forcing an exit')
            sys.exit(0)
        else:
            mpstate.status.exit = True

    # Listen for kill signals to cleanly shutdown modules
    fatalsignals = [signal.SIGTERM]
    try:
        fatalsignals.append(signal.SIGQUIT)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)
    except Exception:
        pass
    if opts.daemon or opts.non_interactive: # SIGINT breaks readline parsing - if we are interactive, just let things die
        fatalsignals.append(signal.SIGINT)

    for sig in fatalsignals:
        signal.signal(sig, quit_handler)

    load_module('link', quiet=True)

    mpstate.settings.source_system = opts.SOURCE_SYSTEM
    mpstate.settings.source_component = opts.SOURCE_COMPONENT

    # open master link
    for mdev in opts.master:
        if mdev.find('?') != -1 or mdev.find('*') != -1:
            for m in glob.glob(mdev):
                if not mpstate.module('link').link_add(m, force_connected=opts.force_connected):
                    sys.exit(1)
        elif not mpstate.module('link').link_add(mdev, force_connected=opts.force_connected):
            sys.exit(1)

    if not opts.master and len(serial_list) == 1:
          print("Connecting to %s" % serial_list[0])
          mpstate.module('link').link_add(serial_list[0].device)
    elif not opts.master and len(serial_list) > 1:
          print("Warning: multiple possible serial ports. Use console GUI or 'link add' to add port, or restart using --master to select a single port")
          #if no display, assume running CLI mode and exit
          if platform.system() != 'Windows' and "DISPLAY" not in os.environ:
              sys.exit(1)
    elif not opts.master:
          wifi_device = '0.0.0.0:14550'
          mpstate.module('link').link_add(wifi_device)


    # open any mavlink output ports
    for port in opts.output:
        mpstate.mav_outputs.append(mavutil.mavlink_connection(port, baud=int(opts.baudrate), input=False))

    if opts.sitl:
        mpstate.sitl_output = mavutil.mavudp(opts.sitl, input=False)

    mpstate.settings.streamrate = opts.streamrate
    mpstate.settings.streamrate2 = opts.streamrate

    if opts.state_basedir is not None:
        mpstate.settings.state_basedir = opts.state_basedir

    msg_period = mavutil.periodic_event(1.0/15)
    heartbeat_period = mavutil.periodic_event(1)
    heartbeat_check_period = mavutil.periodic_event(0.33)

    mpstate.input_queue = Queue.Queue()
    mpstate.input_count = 0
    mpstate.empty_input_count = 0
    if opts.setup:
        mpstate.rl.set_prompt("")

    # call this early so that logdir is setup based on --aircraft
    (mpstate.status.logdir, logpath_telem, logpath_telem_raw) = log_paths()

    for module in opts.load_module:
        modlist = module.split(',')
        for mod in modlist:
            process_stdin('module load %s' % (mod))

    if not opts.setup:
        # some core functionality is in modules
        standard_modules = opts.default_modules.split(',')
        for m in standard_modules:
            if m:
                load_module(m, quiet=True)

    if platform.system() != 'Windows':
        if opts.console:
            process_stdin('module load console')
    else:
        # default to having console on windows
        if opts.console or not opts.no_console:
            process_stdin('module load console')

    if opts.map:
        process_stdin('module load map')

    start_scripts = []
    if not opts.setup:
        if 'HOME' in os.environ:
            start_scripts.append(os.path.join(os.environ['HOME'], ".mavinit.scr"))
        start_script = mp_util.dot_mavproxy("mavinit.scr")
        start_scripts.append(start_script)
    if (mpstate.settings.state_basedir is not None and
        opts.aircraft is not None):
        start_script = os.path.join(mpstate.settings.state_basedir, opts.aircraft, "mavinit.scr")
        start_scripts.append(start_script)
    for start_script in start_scripts:
        if os.path.exists(start_script):
            print("Running script (%s)" % (start_script))
            run_script(start_script)

    if opts.aircraft is not None:
        start_script = os.path.join(opts.aircraft, "mavinit.scr")
        if os.path.exists(start_script):
            run_script(start_script)
        else:
            print("no script %s" % start_script)

    if opts.cmd is not None:
        for cstr in opts.cmd:
            cmds = cstr.split(';')
            for c in cmds:
                process_stdin(c)

    if opts.profile:
        import yappi    # We do the import here so that we won't barf if run normally and yappi not available
        yappi.start()

    # log all packets from the master, for later replay
    open_telemetry_logs(logpath_telem, logpath_telem_raw)

    # run main loop as a thread
    mpstate.status.thread = threading.Thread(target=main_loop, name='main_loop')
    mpstate.status.thread.daemon = True
    mpstate.status.thread.start()

    # use main program for input. This ensures the terminal cleans
    # up on exit
    while (mpstate.status.exit != True):
        try:
            if opts.daemon or opts.non_interactive:
                time.sleep(0.1)
            else:
                input_loop()
        except KeyboardInterrupt:
            if mpstate.settings.requireexit:
                print("Interrupt caught.  Use 'exit' to quit MAVProxy.")

                #Just lost the map and console, get them back:
                for (m,pm) in mpstate.modules:
                    if m.name in ["map", "console"]:
                        if hasattr(m, 'unload'):
                            try:
                                m.unload()
                            except Exception:
                                pass
                        reload(m)
                        m.init(mpstate)

            else:
                mpstate.status.exit = True
                sys.exit(1)

    if opts.profile:
        yappi.get_func_stats().print_all()
        yappi.get_thread_stats().print_all()

    #this loop executes after leaving the above loop and is for cleanup on exit
    for (m,pm) in mpstate.modules:
        if hasattr(m, 'unload'):
            print("Unloading module %s" % m.name)
            m.unload()

    sys.exit(1)



#!/usr/bin/python
'''
Wrapper for the SRTM module (srtm.py)
It will grab the altitude of a long,lat pair from the SRTM database
Created by Stephen Dade (stephen_dade@hotmail.com)
'''

import os
import sys
import time

import numpy

from MAVProxy.modules.mavproxy_map import srtm

class ElevationModel():
    '''Elevation Model. Only SRTM for now'''

    def __init__(self, database='srtm', offline=0, debug=False):
        '''Use offline=1 to disable any downloading of tiles, regardless of whether the
        tile exists'''
        self.database = database
        if self.database == 'srtm':
            self.downloader = srtm.SRTMDownloader(offline=offline, debug=debug)
            self.downloader.loadFileList()
            self.tileDict = dict()

        '''Use the Geoscience Australia database instead - watch for the correct database path'''
        if self.database == 'geoscience':
            from MAVProxy.modules.mavproxy_map import GAreader
            self.mappy = GAreader.ERMap()
            self.mappy.read_ermapper(os.path.join(os.environ['HOME'], './Documents/Elevation/Canberra/GSNSW_P756demg'))

    def GetElevation(self, latitude, longitude, timeout=0):
        '''Returns the altitude (m ASL) of a given lat/long pair, or None if unknown'''
        if latitude is None or longitude is None:
            return None
        if self.database == 'srtm':
            TileID = (numpy.floor(latitude), numpy.floor(longitude))
            if TileID in self.tileDict:
                alt = self.tileDict[TileID].getAltitudeFromLatLon(latitude, longitude)
            else:
                tile = self.downloader.getTile(numpy.floor(latitude), numpy.floor(longitude))
                if tile == 0:
                    if timeout > 0:
                        t0 = time.time()
                        while time.time() < t0+timeout and tile == 0:
                            tile = self.downloader.getTile(numpy.floor(latitude), numpy.floor(longitude))
                            if tile == 0:
                                time.sleep(0.1)
                if tile == 0:
                    return None
                self.tileDict[TileID] = tile
                alt = tile.getAltitudeFromLatLon(latitude, longitude)
        if self.database == 'geoscience':
             alt = self.mappy.getAltitudeAtPoint(latitude, longitude)
        return alt


if __name__ == "__main__":

    from optparse import OptionParser
    parser = OptionParser("mp_elevation.py [options]")
    parser.add_option("--lat", type='float', default=-35.052544, help="start latitude")
    parser.add_option("--lon", type='float', default=149.509165, help="start longitude")
    parser.add_option("--database", type='string', default='srtm', help="elevation database")
    parser.add_option("--debug", action='store_true', help="enabled debugging")

    (opts, args) = parser.parse_args()

    EleModel = ElevationModel(opts.database, debug=opts.debug)

    lat = opts.lat
    lon = opts.lon

    '''Do a few lat/long pairs to demonstrate the caching
    Note the +0.000001 to the time. On faster PCs, the two time periods
    may in fact be equal, so we add a little extra time on the end to account for this'''
    t0 = time.time()
    alt = EleModel.GetElevation(lat, lon, timeout=10)
    if alt is None:
        print("Tile not available")
        sys.exit(1)
    t1 = time.time()+.000001
    print("Altitude at (%.6f, %.6f) is %u m. Pulled at %.1f FPS" % (lat, lon, alt, 1/(t1-t0)))

    lat = opts.lat+0.001
    lon = opts.lon+0.001
    t0 = time.time()
    alt = EleModel.GetElevation(lat, lon, timeout=10)
    t1 = time.time()+.000001
    print("Altitude at (%.6f, %.6f) is %u m. Pulled at %.1f FPS" % (lat, lon, alt, 1/(t1-t0)))

    lat = opts.lat-0.001
    lon = opts.lon-0.001
    t0 = time.time()
    alt = EleModel.GetElevation(lat, lon, timeout=10)
    t1 = time.time()+.000001
    print("Altitude at (%.6f, %.6f) is %u m. Pulled at %.1f FPS" % (lat, lon, alt, 1/(t1-t0)))






"""
setpos command for SET_POSITION_TARGET_LOCAL_NED
"""

import math
import time

from MAVProxy.modules.lib import mp_module
from pymavlink import mavutil
from MAVProxy.modules.lib import mp_settings


class SetPosModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(SetPosModule, self).__init__(mpstate, "SetPos", "SetPos", public=False)
        self.add_command('setpos', self.cmd_setpos, "set local pos")
        self.add_command('hop', self.cmd_hop, "hop position")
        self.hop = None
        self.hop_stage = 0
        self.hop_last_time = time.time()
        self.hop_settings = mp_settings.MPSettings(
            [('height', float, 0.7),
             ('takeoff_delay', float, 4.0),
             ('move_delay', float, 4.0)])
        self.add_command('hop', self.cmd_hop, 'HOP control',
                         ["set (HOPSETTING)"])
        self.add_completion_function('(HOPSETTING)',
                                     self.hop_settings.completion)

    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        pass

    def cmd_setpos(self, args):
        '''set local position'''
        if len(args) < 3:
            print("Usage: setpos dX dY dZ dYaw")
            return
        dX = float(args[0])
        dY = float(args[1])
        dZ = float(args[2])
        if len(args) >= 4:
            dYaw = math.radians(float(args[3]))
        else:
            dYaw = 0.0
        self.master.mav.set_position_target_local_ned_send(
            0, # timestamp
            self.target_system, # target system_id
            self.target_component, # target component id
            mavutil.mavlink.MAV_FRAME_BODY_OFFSET_NED,
            0b1111101111111000, # mask specifying use-only-x-y-z-yaw
            dX, # x
            dY, # y
            -dZ,# z
            0, # vx
            0, # vy
            0, # vz
            0, # afx
            0, # afy
            0, # afz
            dYaw, # yaw
            0, # yawrate
            )

    def cmd_hop(self, args):
        '''start a hop'''
        if len(args) > 0 and args[0] == "set":
            self.hop_settings.command(args[1:])
            return
        if len(args) < 2:
            print("Usage: hop dX dY dYaw")
            return
        dX = float(args[0])
        dY = float(args[1])
        if len(args) > 2:
            dYaw = float(args[2])
        else:
            dYaw = 0
        self.hop_stage = 0
        self.hop_last_time = time.time()
        self.hop = [dX, dY, dYaw]

    def idle_task(self):
        '''run commands when idle'''
        if self.hop is None:
            return
        if time.time() - self.hop_last_time < 0.2:
            return
        if self.status.flightmode != 'GUIDED':
            self.module('mode').cmd_mode(['GUIDED'])
            self.hop_last_time = time.time()
            return
        if not self.master.motors_armed():
            self.module('arm').cmd_arm(['throttle'])
            self.hop_last_time = time.time()
            return
        now = time.time()
        if self.hop_stage == 0:
            self.module('cmdlong').cmd_takeoff([self.hop_settings.height])
            self.hop_last_time = time.time()
            self.hop_stage += 1
        if self.hop_stage == 1:
            if now - self.hop_last_time < self.hop_settings.takeoff_delay:
                return
            dX = self.hop[0]
            dY = self.hop[1]
            dYaw = self.hop[2]
            self.cmd_setpos([dX, dY, 0, dYaw])
            self.hop_last_time = time.time()
            self.hop_stage += 1
        if self.hop_stage == 2:
            if now - self.hop_last_time < self.hop_settings.move_delay:
                return
            self.module('mode').cmd_mode(['LAND'])
            self.hop = None


def init(mpstate):
    '''initialise module'''
    return SetPosModule(mpstate)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
slipmap based on mp_tile
Andrew Tridgell
June 2012
'''

import functools
import math
import os, sys
import time
import cv2
import numpy as np

from MAVProxy.modules.mavproxy_map import mp_elevation
from MAVProxy.modules.mavproxy_map import mp_tile
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import win_layout
from MAVProxy.modules.lib import multiproc
from MAVProxy.modules.mavproxy_map.mp_slipmap_util import *


class MPSlipMap():
    '''
    a generic map viewer widget for use in mavproxy
    '''
    def __init__(self,
                 title='SlipMap',
                 lat=-35.362938,
                 lon=149.165085,
                 width=800,
                 height=600,
                 ground_width=1000,
                 tile_delay=0.3,
                 service="MicrosoftSat",
                 max_zoom=19,
                 debug=False,
                 brightness=0,
                 elevation=False,
                 download=True,
                 show_flightmode_legend=True,
                 timelim_pipe=None):

        self.lat = lat
        self.lon = lon
        self.width = width
        self.height = height
        self.ground_width = ground_width
        self.download = download
        self.service = service
        self.tile_delay = tile_delay
        self.debug = debug
        self.max_zoom = max_zoom
        self.elevation = elevation
        self.oldtext = None
        self.brightness = brightness
        self.legend = show_flightmode_legend
        self.timelim_pipe = timelim_pipe

        self.drag_step = 10

        self.title = title
        self.app_ready = multiproc.Event()
        self.event_queue = multiproc.Queue()
        self.object_queue = multiproc.Queue()
        self.close_window = multiproc.Semaphore()
        self.close_window.acquire()
        self.child = multiproc.Process(target=self.child_task)
        self.child.start()
        self._callbacks = set()

        # ensure the map application is ready before returning
        if not self._wait_ready(timeout=5.0):
            raise Exception("map not ready")

    def child_task(self):
        '''child process - this holds all the GUI elements'''
        mp_util.child_close_fds()

        from MAVProxy.modules.lib import wx_processguard
        from MAVProxy.modules.lib.wx_loader import wx
        from MAVProxy.modules.mavproxy_map.mp_slipmap_ui import MPSlipMapFrame

        state = self

        self.mt = mp_tile.MPTile(download=self.download,
                                 service=self.service,
                                 tile_delay=self.tile_delay,
                                 debug=self.debug,
                                 max_zoom=self.max_zoom)
        state.layers = {}
        state.info = {}
        state.need_redraw = True

        self.app = wx.App(False)
        self.app.SetExitOnFrameDelete(True)
        self.app.frame = MPSlipMapFrame(state=self)
        self.app.frame.Show()
        self.app_ready.set()
        self.app.MainLoop()

    def close(self):
        '''close the window'''
        self.close_window.release()
        count=0
        while self.child.is_alive() and count < 30: # 3 seconds to die...
            time.sleep(0.1) #?
            count+=1

        if self.child.is_alive():
            self.child.terminate()

        self.child.join()

    def is_alive(self):
        '''check if graph is still going'''
        return self.child.is_alive()

    def add_object(self, obj):
        '''add or update an object on the map'''
        self.object_queue.put(obj)

    def remove_object(self, key):
        '''remove an object on the map by key'''
        self.object_queue.put(SlipRemoveObject(key))

    def set_zoom(self, ground_width):
        '''set ground width of view'''
        self.object_queue.put(SlipZoom(ground_width))

    def set_center(self, lat, lon):
        '''set center of view'''
        self.object_queue.put(SlipCenter((lat,lon)))

    def set_follow(self, enable):
        '''set follow on/off'''
        self.object_queue.put(SlipFollow(enable))

    def set_follow_object(self, key, enable):
        '''set follow on/off on an object'''
        self.object_queue.put(SlipFollowObject(key, enable))
        
    def hide_object(self, key, hide=True):
        '''hide an object on the map by key'''
        self.object_queue.put(SlipHideObject(key, hide))

    def set_position(self, key, latlon, layer='', rotation=0, label=None, colour=None):
        '''move an object on the map'''
        self.object_queue.put(SlipPosition(key, latlon, layer, rotation, label, colour))

    def event_queue_empty(self):
        '''return True if there are no events waiting to be processed'''
        return self.event_queue.empty()

    def set_layout(self, layout):
        '''set window layout'''
        self.object_queue.put(layout)
    
    def get_event(self):
        '''return next event or None'''
        if self.event_queue.empty():
            return None
        evt = self.event_queue.get()
        while isinstance(evt, win_layout.WinLayout):
            win_layout.set_layout(evt, self.set_layout)
            if self.event_queue.empty():
                return None
            evt = self.event_queue.get()
        return evt

    def add_callback(self, callback):
        '''add a callback for events from the map'''
        self._callbacks.add(callback)

    def check_events(self):
        '''check for events, calling registered callbacks as needed'''
        while not self.event_queue_empty():
            event = self.get_event()
            for callback in self._callbacks:
                callback(event)

    def icon(self, filename):
        '''load an icon from the data directory'''
        return mp_tile.mp_icon(filename)

    def _wait_ready(self, timeout):
        '''Wait at most timeout for the application to be ready

        Param
        -----
        timeout: float
            timeout in seconds

        Returns True if the map is ready, False if the timeout is reached.
        '''
        start_time = time.time()
        while time.time() - start_time < timeout:
            if self.app_ready.is_set():
                return True
            time.sleep(0.1)
        return False

if __name__ == "__main__":
    multiproc.freeze_support()
    import time

    from optparse import OptionParser
    parser = OptionParser("mp_slipmap.py [options]")
    parser.add_option("--lat", type='float', default=-26.582218, help="start latitude")
    parser.add_option("--lon", type='float', default=151.840113, help="start longitude")
    parser.add_option("--service", default="MicrosoftSat", help="tile service")
    parser.add_option("--offline", action='store_true', default=False, help="no download")
    parser.add_option("--delay", type='float', default=0.3, help="tile download delay")
    parser.add_option("--max-zoom", type='int', default=19, help="maximum tile zoom")
    parser.add_option("--debug", action='store_true', default=False, help="show debug info")
    parser.add_option("--boundary", default=None, help="show boundary")
    parser.add_option("--mission", default=[], action='append', help="show mission")
    parser.add_option("--thumbnail", default=None, help="show thumbnail")
    parser.add_option("--icon", default=None, help="show icon")
    parser.add_option("--flag", default=[], type='str', action='append', help="flag positions")
    parser.add_option("--grid", default=False, action='store_true', help="add a UTM grid")
    parser.add_option("--elevation", action='store_true', default=False, help="show elevation information")
    parser.add_option("--verbose", action='store_true', default=False, help="show mount actions")
    (opts, args) = parser.parse_args()

    sm = MPSlipMap(lat=opts.lat,
                   lon=opts.lon,
                   download=not opts.offline,
                   service=opts.service,
                   debug=opts.debug,
                   max_zoom=opts.max_zoom,
                   elevation=opts.elevation,
                   tile_delay=opts.delay)

    if opts.boundary:
        boundary = mp_util.polygon_load(opts.boundary)
        sm.add_object(SlipPolygon('boundary', boundary, layer=1, linewidth=2, colour=(0,255,0)))

    if opts.mission:
        from pymavlink import mavwp
        for file in opts.mission:
            wp = mavwp.MAVWPLoader()
            wp.load(file)
            boundary = wp.polygon()
            sm.add_object(SlipPolygon('mission-%s' % file, boundary, layer=1, linewidth=1, colour=(255,255,255)))

    if opts.grid:
        sm.add_object(SlipGrid('grid', layer=3, linewidth=1, colour=(255,255,0)))

    if opts.thumbnail:
        thumb = cv2.imread(opts.thumbnail)
        sm.add_object(SlipThumbnail('thumb', (opts.lat,opts.lon), layer=1, img=thumb, border_width=2, border_colour=(255,0,0)))

    if opts.icon:
        icon = cv2.imread(opts.icon)
        sm.add_object(SlipIcon('icon', (opts.lat,opts.lon), icon, layer=3, rotation=90, follow=True))
        sm.set_position('icon', mp_util.gps_newpos(opts.lat,opts.lon, 180, 100), rotation=45)
        sm.add_object(SlipInfoImage('detail', icon))
        sm.add_object(SlipInfoText('detail text', 'test text'))

    for flag in opts.flag:
        (lat,lon) = flag.split(',')
        icon = sm.icon('flag.png')
        sm.add_object(SlipIcon('icon - %s' % str(flag), (float(lat),float(lon)), icon, layer=3, rotation=0, follow=False))

    while sm.is_alive():
        while not sm.event_queue_empty():
            obj = sm.get_event()
            if not opts.verbose:
                continue
            if isinstance(obj, SlipMouseEvent):
                print("Mouse event at %s (X/Y=%u/%u) for %u objects" % (obj.latlon,
                                                                        obj.event.X, obj.event.Y,
                                                                        len(obj.selected)))
            if isinstance(obj, SlipKeyEvent):
                print("Key event at %s for %u objects" % (obj.latlon, len(obj.selected)))
        time.sleep(0.1)



#!/usr/bin/env python
'''
 test follow-me options in ArduPilot
 Andrew Tridgell
 September 2016
'''

import sys, os, time, math
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.mavproxy_map import mp_slipmap
from pymavlink import mavutil
if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

class FollowTestModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(FollowTestModule, self).__init__(mpstate, "followtest", "followtest module")
        self.add_command('followtest', self.cmd_followtest, "followtest control",
                         ['set (FOLLOWSETTING)'])
        self.follow_settings = mp_settings.MPSettings([("radius", float, 100.0),
                                                       ("altitude", float, 50.0),
                                                       ("speed", float, 10.0),
                                                       ("type", str, 'guided'),
                                                       ("vehicle_throttle", float, 0.5),
                                                       ("disable_msg", bool, False)])
        self.add_completion_function('(FOLLOWSETTING)', self.follow_settings.completion)
        self.target_pos = None
        self.last_update = 0
        self.circle_dist = 0
        
    def cmd_followtest(self, args):
        '''followtest command parser'''
        usage = "usage: followtest <set>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            self.follow_settings.command(args[1:])
        else:
            print(usage)

    def update_target(self, time_boot_ms):
        '''update target on map'''
        if not self.mpstate.map:
            # don't draw if no map
            return
        if not 'HOME_POSITION' in self.master.messages:
            return

        home_position = self.master.messages['HOME_POSITION']

        now = time_boot_ms * 1.0e-3
        dt = now - self.last_update
        if dt < 0:
            dt = 0
        self.last_update = now

        self.circle_dist += dt * self.follow_settings.speed

        # assume a circle for now
        circumference = math.pi * self.follow_settings.radius * 2
        rotations = math.fmod(self.circle_dist, circumference) / circumference
        angle = math.pi * 2 * rotations
        self.target_pos = mp_util.gps_newpos(home_position.latitude*1.0e-7,
                                             home_position.longitude*1.0e-7,
                                             math.degrees(angle),
                                             self.follow_settings.radius)

        icon = self.mpstate.map.icon('camera-small-red.png')
        (lat, lon) = (self.target_pos[0], self.target_pos[1])
        self.mpstate.map.add_object(mp_slipmap.SlipIcon('followtest',
                                                        (lat, lon),
                                                        icon, layer='FollowTest', rotation=0, follow=False))
        

    def idle_task(self):
        '''update vehicle position'''
        pass


    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if not self.mpstate.map:
            # don't draw if no map
            return

        if m.get_type() != 'GLOBAL_POSITION_INT':
            return
        self.update_target(m.time_boot_ms)

        if self.target_pos is None:
            return

        if self.follow_settings.disable_msg:
            return

        if self.follow_settings.type == 'guided':
            # send normal guided mode packet
            self.master.mav.mission_item_int_send(self.settings.target_system,
                                                  self.settings.target_component,
                                                  0,
                                                  self.module('wp').get_default_frame(),
                                                  mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
                                                  2, 0, 0, 0, 0, 0,
                                                  int(self.target_pos[0]*1.0e7), int(self.target_pos[1]*1.0e7),
                                                  self.follow_settings.altitude)

        elif self.follow_settings.type == 'yaw':
            # display yaw from vehicle to target
            vehicle = (m.lat*1.0e-7, m.lon*1.0e-7)
            vehicle_yaw = math.degrees(self.master.field('ATTITUDE', 'yaw', 0))
            target_bearing = mp_util.gps_bearing(vehicle[0], vehicle[1], self.target_pos[0], self.target_pos[1])
            # wrap the angle from -180 to 180 thus commanding the vehicle to turn left or right
            # note its in centi-degrees so *100
            relyaw = mp_util.wrap_180(target_bearing - vehicle_yaw) * 100

            self.master.mav.command_long_send(self.settings.target_system,
                                                  self.settings.target_component,
                                                  mavutil.mavlink.MAV_CMD_NAV_SET_YAW_SPEED, 0,
                                                  relyaw,
                                                  self.follow_settings.vehicle_throttle,
                                                  0, 0, 0, 0, 0)

def init(mpstate):
    '''initialise module'''
    return FollowTestModule(mpstate)



#!/usr/bin/env python
'''
silvus radio support
To use add this in mavinit.scr for vehicle:

 module load silvus
 silvus set gnd_ip 172.20.167.5
 silvus set air_ip 172.20.168.66
 silvus set nmea_ip 172.20.167.5
 silvus set nmea_port 11234 # UDP port for NMEA input to radio

thanks to Felix from Amber Technologies for the code this came from
'''

import time
import datetime
import socket
import requests
import threading
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from pymavlink import mavutil

class SilvusModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(SilvusModule, self).__init__(mpstate, "Silvus", "Silvus output")
        # filter_dist is distance in metres
        self.silvus_settings = mp_settings.MPSettings([("gnd_ip", str, ""),
                                                       ("air_ip", str, ""),
                                                       ("nmea_ip", str, ""),
                                                       ("nmea_port", int, -1),
                                                       ('log_dt', float, 1.0),
                                                       ('debug', int, 0),
                                                       ])
        self.add_completion_function('(SILVUSSETTING)',
                                     self.silvus_settings.completion)
        self.add_command('silvus', self.cmd_silvus, "silvus control",
                         ["status", "set (SILVUSSETTING)"])
        self.last_nmea_send = time.time()
        self.last_log_time = time.time()

        self.thread = threading.Thread(target=self.thread_loop)
        self.thread.start()
        self.values = {}

    def cmd_silvus(self, args):
        '''silvus commands'''
        if args[0] == "set":
            self.silvus_settings.command(args[1:])
        elif args[0] == "status":
            self.cmd_status()

    def nmea_checkstr(self, msg):
        d = msg[1:]
        cs = 0
        for i in d:
            cs ^= ord(i)
        return "*%02X\r\n" % cs

    def send_nmea(self):
        '''send a NMEA packet to a radio, so the radio knows its position for logging and display'''

        if not self.silvus_settings.nmea_ip or self.silvus_settings.nmea_port <= 0:
            print("no setup")
            return

        now_time = time.time()
        if now_time - self.last_nmea_send < 1.0:
            return
        self.last_nmea_send = now_time

        gps = self.master.messages.get('GPS_RAW_INT', None)
        if gps is None or gps.fix_type < 3:
            return

        lat = gps.lat * 1.0e-7
        lon = gps.lon * 1.0e-7
        alt = gps.alt * 1.0e-3
        nsat = gps.satellites_visible
        hdop = gps.eph/100.0
        fix = gps.fix_type
        speed = ((gps.vel/100.0)/1852.0)*3600 # knots
        course = gps.cog/100.0

        output = ""
        utc_sec = now_time
        tm_t = time.gmtime(utc_sec)
        dstr = "%02d%02d%02d" % (tm_t.tm_mday, tm_t.tm_mon, tm_t.tm_year % 100)
        subsecs = utc_sec - int(utc_sec);
        tstr = "%02d%02d%05.3f" % (tm_t.tm_hour, tm_t.tm_min, tm_t.tm_sec + subsecs)
        deg = abs(lat)
        minutes = (deg - int(deg))*60
        latstr = "%02d%08.5f,%c" % (int(deg), minutes, 'S' if lat < 0 else 'N')
        deg = abs(lon)
        minutes = (deg - int(deg))*60
        lonstr = "%03d%08.5f,%c" % (int(deg), minutes, 'W' if lon < 0 else 'E')

        gga = "$GPGGA,%s,%s,%s,%01d,%02d,%04.1f,%07.2f,M,0.0,M,," % (tstr, latstr, lonstr, fix, nsat, hdop, alt)
        output = output + gga + self.nmea_checkstr(gga)

        rmc = "$GPRMC,%s,%s,%s,%s,%.2f,%.2f,%s,," % (tstr, fix, latstr, lonstr, speed, course, dstr)
        output = output + rmc + self.nmea_checkstr(rmc)

        if self.silvus_settings.debug > 1:
            print("NMEA: %s" % output)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.sendto(output.encode("UTF-8"), (self.silvus_settings.nmea_ip, self.silvus_settings.nmea_port))
        except Exception as ex:
            print("Silvus NMEA send fail: %s" % ex)

    def url(self, nodeip, api):
        return "http://%s/%s" % (nodeip, api)

    # Get Frequency == freq (nodeip)
    def get_freq(self, nodeip):
        data = '{"jsonrpc":"2.0","method":"freq","id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        freql = (response.json()["result"])
        freq = int(freql[0])
        return freq

    # Get NoiseLevel == noise(nodeip)
    def get_noise(self, nodeip):
        data = '{"jsonrpc":"2.0","method":"noise_level","id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        noisel = (response.json()["result"])
        return int(noisel[0])

    # Get Neighbor RSSI == nbr_rssi(nodeip, localnode)
    def get_rssi(self, nodeip, remotenode):
        data = '{"jsonrpc":"2.0","method":"nbr_rssi","params":["' + remotenode + '"],"id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        nbr_rssi = (response.json()["result"])
        return nbr_rssi

    # Get network Status, output as an array
    def network_status(self, nodeip):
        data = '{"jsonrpc":"2.0","method":"network_status","id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        netstat = (response.json()["result"])
        return netstat

    # Get GPS status
    def get_gps_state(self, nodeip):
        data = '{"jsonrpc":"2.0","method":"gps_mode","id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        gpsstat = (response.json()["result"])
        return gpsstat

    # Get max throughput between nodes
    def get_throughput(self, nodeip, remotenode):
        data = '{"jsonrpc":"2.0","method":"link_throughput","params":["' + remotenode + '", "1"],"id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        nbr_tp = (response.json()["result"])
        nbr_tp = (nbr_tp)[0]
        return nbr_tp

    # Returns the TX MCS
    def get_neighbor_mcs(self, nodeip, remotenode):
        data = '{"jsonrpc":"2.0","method":"nbr_mcs","params":["' + remotenode + '"],"id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        nbr_mcs = (response.json()["result"])
        nbr_mcs = (nbr_mcs)[0]
        # print(data)
        return nbr_mcs

    # Returns the RX MCS
    def get_neighbor_mcs_rx(self, nodeip, remotenode):
        data = '{"jsonrpc":"2.0","method":"nbr_mcs_rx","params":["' + remotenode + '"],"id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        nbr_mcs_rx = (response.json()["result"])
        nbr_mcs_rx = (nbr_mcs_rx)[0]
        # print(data)
        return nbr_mcs_rx

    # Get GPS coords, output as an array
    def get_gps_coords(self, nodeip):
        data = '{"jsonrpc":"2.0","method":"gps_coordinates","id":"sbkb5u0c"}'
        response = requests.post(self.url(nodeip, 'streamscape_api'), data=data)
        coords = (response.json()["result"])
        return coords

    # convert last 2 octets of IP to a nodeid
    def ip_to_id(self, ip):
        ip = ip.split('.')
        return ((int(ip[2]) * 256) + int(ip[3]))

    def send_named_float(self, name, value):
        '''inject a NAMED_VALUE_FLOAT into the local master input, so it becomes available
           for graphs, logging and status command'''

        # use the ATTITUDE message for srcsystem and time stamps
        att = self.master.messages.get('ATTITUDE',None)
        if att is None:
            return
        msec = att.time_boot_ms
        ename = name.encode('ASCII')
        if len(ename) < 10:
            ename += bytes([0] * (10-len(ename)))
        m = self.master.mav.named_value_float_encode(msec, bytearray(ename), value)
        #m.name = ename
        m.pack(self.master.mav)
        m._header.srcSystem = att._header.srcSystem
        m._header.srcComponent = mavutil.mavlink.MAV_COMP_ID_TELEMETRY_RADIO
        m.name = name
        self.mpstate.module('link').master_callback(m, self.master)
        if self.silvus_settings.debug > 0:
            print(m)

    def get_radio_data(self):
        now = time.time()
        if now - self.last_log_time < self.silvus_settings.log_dt:
            return
        self.last_log_time = now

        localip = self.silvus_settings.gnd_ip
        remoteip = self.silvus_settings.air_ip

        if len(localip.split('.')) != 4:
            return
        if len(remoteip.split('.')) != 4:
            return

        localnode = str(self.ip_to_id(localip))
        remotenode = str(self.ip_to_id(remoteip))

        self.values['TXMCS'] = float(self.get_neighbor_mcs(localip, remotenode))
        self.values['RXMCS'] = float(self.get_neighbor_mcs_rx(localip, remotenode))
        rssi = self.get_rssi(localip, remotenode)
        self.values['TXRSSI1'] = float(rssi[0])
        self.values['TXRSSI2'] = float(rssi[1])
        self.values['TXRSSI3'] = float(rssi[2])
        self.values['TXRSSI4'] = float(rssi[3])
        rssi = self.get_rssi(remoteip, localnode)
        self.values['RXRSSI1'] = float(rssi[0])
        self.values['RXRSSI2'] = float(rssi[1])
        self.values['RXRSSI3'] = float(rssi[2])
        self.values['RXRSSI4'] = float(rssi[3])
        self.values['LOCNSE'] = float(self.get_noise(localip))
        self.values['REMNSE'] = float(self.get_noise(remoteip))
        self.values['LINKSNR'] = float(self.network_status(localip)[2])
        self.values['LOCTPUT'] = float(self.get_throughput(localip, remotenode))
        self.values['REMTPUT'] = float(self.get_throughput(remoteip, localnode))

        for f in self.values:
            self.send_named_float('SR_' + f, self.values[f])

    def cmd_status(self):
        for f in sorted(self.values.keys()):
            print("%20s %.1f" % (f, self.values[f]))

    def thread_loop(self):
        while True:
            time.sleep(self.silvus_settings.log_dt)
            try:
                self.send_nmea()
                self.get_radio_data()
            except Exception as ex:
                if self.silvus_settings.debug > 0:
                    print(ex)


def init(mpstate):
    '''initialise module'''
    return SilvusModule(mpstate)



"""
send NTRIP data to flight controller
"""

import time

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import ntrip
from MAVProxy.modules.lib import mp_settings


class NtripModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(NtripModule, self).__init__(mpstate, "ntrip", "ntrip", public=False)
        self.ntrip_settings = mp_settings.MPSettings(
            [('caster', str, None),
             ('port', int, 2101),
             ('username', str, 'IBS'),
             ('password', str, 'IBS'),
             ('mountpoint', str, None),
             ('logfile', str, None),
             ('sendalllinks', bool, False),
             ('sendmul', int, 1)])
        self.add_command('ntrip', self.cmd_ntrip, 'NTRIP control',
                         ["<status>",
                          "<start>",
                          "<stop>",
                          "set (NTRIPSETTING)"])
        self.add_completion_function('(NTRIPSETTING)',
                                     self.ntrip_settings.completion)
        self.pos = None
        self.pkt_count = 0
        self.last_pkt = None
        self.last_restart = None
        self.last_rate = None
        self.rate_total = 0
        self.ntrip = None
        self.start_pending = False
        self.rate = 0
        self.logfile = None
        self.id_counts = {}
        self.last_by_id = {}

    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        if msg.get_type() in ['GPS_RAW_INT', 'GPS2_RAW']:
            if msg.fix_type >= 3:
                self.pos = (msg.lat*1.0e-7, msg.lon*1.0e-7, msg.alt*1.0e-3)

    def log_rtcm(self, data):
        '''optionally log rtcm data'''
        if self.ntrip_settings.logfile is None:
            return
        if self.logfile is None:
            self.logfile = open(self.ntrip_settings.logfile, 'wb')
        if self.logfile is not None:
            self.logfile.write(data)

    def idle_task(self):
        '''called on idle'''
        if self.start_pending and self.ntrip is None and self.pos is not None:
            self.cmd_start()
        if self.ntrip is None:
            return
        data = self.ntrip.read()
        if data is None:
            now = time.time()
            if (self.last_pkt is not None and
                now - self.last_pkt > 15 and
                (self.last_restart is None or now - self.last_restart > 30)):
                print("NTRIP restart")
                self.ntrip = None
                self.start_pending = True
                self.last_restart = now
            return
        if time.time() - self.ntrip.dt_last_gga_sent > 2:
            self.ntrip.setPosition(self.pos[0], self.pos[1])
            self.ntrip.send_gga()
        self.log_rtcm(data)

        rtcm_id = self.ntrip.get_ID()
        if not rtcm_id in self.id_counts:
            self.id_counts[rtcm_id] = 0
            self.last_by_id[rtcm_id] = data[:]
        self.id_counts[rtcm_id] += 1

        blen = len(data)
        if blen > 4*180:
            # can't send this with GPS_RTCM_DATA
            return
        total_len = blen
        self.rate_total += blen * self.ntrip_settings.sendmul

        if blen > 180:
            flags = 1 # fragmented
        else:
            flags = 0
        # add in the sequence number
        flags |= (self.pkt_count & 0x1F) << 3

        fragment = 0
        while blen > 0:
            send_data = bytearray(data[:180])
            frag_len = len(send_data)
            data = data[frag_len:]
            if frag_len < 180:
                send_data.extend(bytearray([0]*(180-frag_len)))
            if self.ntrip_settings.sendalllinks:
                links = self.mpstate.mav_master
            else:
                links = [self.master]
            for link in links:
                for d in range(self.ntrip_settings.sendmul):
                    link.mav.gps_rtcm_data_send(flags | (fragment<<1), frag_len, send_data)
            fragment += 1
            blen -= frag_len
        self.pkt_count += 1

        now = time.time()
        if now - self.last_rate > 1:
            dt = now - self.last_rate
            rate_now = self.rate_total / float(dt)
            self.rate = 0.9 * self.rate + 0.1 * rate_now
            self.last_rate = now
            self.rate_total = 0
        self.last_pkt = now

    def cmd_ntrip(self, args):
        '''ntrip command handling'''
        if len(args) <= 0:
            print("Usage: ntrip <start|stop|status|set>")
            return
        if args[0] == "start":
            self.cmd_start()
        if args[0] == "stop":
            self.ntrip = None
            self.start_pending = False
        elif args[0] == "status":
            self.ntrip_status()
        elif args[0] == "set":
            self.ntrip_settings.command(args[1:])

    def ntrip_status(self):
        '''show ntrip status'''
        now = time.time()
        if self.ntrip is None:
            print("ntrip: Not started")
            return
        elif self.last_pkt is None:
            print("ntrip: no data")
            return
        frame_size = 0
        for id in sorted(self.id_counts.keys()):
            print(" %4u: %u (len %u)" % (id, self.id_counts[id], len(self.last_by_id[id])))
            frame_size += len(self.last_by_id[id])
        print("ntrip: %u packets, %.1f bytes/sec last %.1fs ago framesize %u" % (self.pkt_count, self.rate, now - self.last_pkt, frame_size))

    def cmd_start(self):
        '''start ntrip link'''
        if self.ntrip_settings.caster is None:
            print("Require caster")
            return
        if self.ntrip_settings.mountpoint is None:
            print("Require mountpoint")
            return
        if self.pos is None:
            print("Start delayed pending position")
            self.start_pending = True
            return
        user = self.ntrip_settings.username + ":" + self.ntrip_settings.password
        self.ntrip = ntrip.NtripClient(user=user,
                                       port=self.ntrip_settings.port,
                                       caster=self.ntrip_settings.caster,
                                       mountpoint=self.ntrip_settings.mountpoint,
                                       lat=self.pos[0],
                                       lon=self.pos[1],
                                       height=self.pos[2])
        print("NTRIP started")
        self.start_pending = False
        self.last_rate = time.time()
        self.rate_total = 0


def init(mpstate):
    '''initialise module'''
    return NtripModule(mpstate)



from pymavlink import mavutil
import fnmatch

miss_cmds = {}
frame_enum = {0: "Abs", 3: "Rel", 10: "AGL"}
frame_enum_rev = {v:k for k,v in frame_enum.items()}

# auto-generate the list of mission commands
for cmd in mavutil.mavlink.enums['MAV_CMD']:
    enum = mavutil.mavlink.enums['MAV_CMD'][cmd]
    name = enum.name
    name = name.replace('MAV_CMD_','')
    if name == 'ENUM_END':
        continue
    miss_cmds[cmd] = name

def cmd_reverse_lookup(command_name):
    '''returns 0 if key not found'''
    for key, value in miss_cmds.items():
        if (value.upper() == command_name.upper()):
            return key
    return 0

# a wildcard map from parameter descriptions to column names. If not found in this map
# then the default "Pn" is used
description_map = [
    ('Empty'             , '-'),
    ('Latitude*'         , 'Lat'),
    ('Longitude*'        , 'Lon'),
    ('Altitude*'         , 'Alt'),
    ('Minimum pitch*'    , 'Pitch'),
    ('Yaw*'              , 'Yaw'),
    ('Desired yaw*'      , 'Yaw'),
    ('Radius*'           , 'Radius'),
    ('Turns*'            , 'Turns'),
    ('Seconds*'          , 'Time(s)'),
    ('Delay in seconds*' , 'Time(s)'),
    ('On / Off*'         , 'Enable'),
    ('Descent / Ascend'  , 'Rate'),
    ('Finish Altitude'   , 'Altitude'),
    ('Distance*'         , 'Distance'),
    ('Mode*'             , 'Mode'),
    ('Custom mode*'      , 'CustomMode'),
    ('Sequence*'         , 'Seq'),
    ('Repeat*'           , 'Repeat'),
    ('Speed type*'       , 'SpeedType'),
    ('Speed*'            , 'Speed'),
    ('Throttle*'         , 'Throttle')
    ]

def make_column_label(command_name, description, default):
    '''try to work out a reasonable column name from parameter description'''
    for (pattern, label) in description_map:
        if fnmatch.fnmatch(description, pattern):
            return label
    return default


def get_column_labels(command_name):
    '''return dictionary of column labels if available'''
    cmd = cmd_reverse_lookup(command_name)
    if cmd == 0:
        return {}
    labels = {}
    enum = mavutil.mavlink.enums['MAV_CMD'][cmd]
    for col in enum.param.keys():
        labels[col] = make_column_label(command_name, enum.param[col], "P%u" % col)
    return labels



#!/usr/bin/env python
'''
control MAVLink2 signing
'''

from pymavlink import mavutil
import time, struct, math, sys

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util

if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

class SigningModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(SigningModule, self).__init__(mpstate, "signing", "signing control", public=True)
        self.add_command('signing', self.cmd_signing, "signing control",
                         ["<setup|remove|disable|key>"])
        self.allow = None

    def cmd_signing(self, args):
        '''handle link commands'''
        usage = "signing: <setup|remove|disable|key> passphrase"
        if len(args) == 0:
            print(usage)
        elif args[0] == 'setup':
            self.cmd_signing_setup(args[1:])
        elif args[0] == 'key':
            self.cmd_signing_key(args[1:])
        elif args[0] == 'disable':
            self.cmd_signing_disable(args[1:])
        elif args[0] == 'remove':
            self.cmd_signing_remove(args[1:])
        else:
            print(usage)

    def passphrase_to_key(self, passphrase):
        '''convert a passphrase to a 32 byte key'''
        import hashlib
        h = hashlib.new('sha256')
        if sys.version_info[0] >= 3:
            passphrase = passphrase.encode('ascii')
        h.update(passphrase)
        return h.digest()

    def cmd_signing_setup(self, args):
        '''setup signing key on board'''
        if len(args) == 0:
            print("usage: signing setup passphrase")
            return
        if not self.master.mavlink20():
            print("You must be using MAVLink2 for signing")
            return
        passphrase = args[0]
        key = self.passphrase_to_key(passphrase)
        secret_key = []
        for b in key:
            if sys.version_info[0] >= 3:
                secret_key.append(b)
            else:
                secret_key.append(ord(b))

        epoch_offset = 1420070400
        now = max(time.time(), epoch_offset)
        initial_timestamp = int((now - epoch_offset)*1e5)
        self.master.mav.setup_signing_send(self.target_system, self.target_component,
                                           secret_key, initial_timestamp)
        print("Sent secret_key")
        self.cmd_signing_key([passphrase])


    def allow_unsigned(self, mav, msgId):
        '''see if an unsigned packet should be allowed'''
        if self.allow is None:
            self.allow = {
                mavutil.mavlink.MAVLINK_MSG_ID_RADIO : True,
                mavutil.mavlink.MAVLINK_MSG_ID_RADIO_STATUS : True
                }
        if msgId in self.allow:
            return True
        if self.settings.allow_unsigned:
            return True
        return False

    def cmd_signing_key(self, args):
        '''set signing key on connection'''
        if len(args) == 0:
            print("usage: signing setup passphrase")
            return
        if not self.master.mavlink20():
            print("You must be using MAVLink2 for signing")
            return
        passphrase = args[0]
        key = self.passphrase_to_key(passphrase)
        self.master.setup_signing(key, sign_outgoing=True, allow_unsigned_callback=self.allow_unsigned)
        print("Setup signing key")

    def cmd_signing_disable(self, args):
        '''disable signing locally'''
        self.master.disable_signing()
        print("Disabled signing")

    def cmd_signing_remove(self, args):
        '''remove signing from server'''
        if not self.master.mavlink20():
            print("You must be using MAVLink2 for signing")
            return
        self.master.mav.setup_signing_send(self.target_system, self.target_component, [0]*32, 0)
        self.master.disable_signing()
        print("Removed signing")

def init(mpstate):
    '''initialise module'''
    return SigningModule(mpstate)



#!/usr/bin/env python
'''gopro control over mavlink for the solo-gimbal

To use this module connect to a Solo with a GoPro installed on the gimbal.
'''

import time, os

from MAVProxy.modules.lib import mp_module
from pymavlink import mavutil

class GoProModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(GoProModule, self).__init__(mpstate, "gopro", "gopro handling")

        self.add_command('gopro', self.cmd_gopro,   'gopro control', [
                                        'status',
                                        'shutter <start|stop>',
                                        'mode <video|camera>',
                                        'power <on|off>'])

    def cmd_gopro(self, args):
        '''gopro commands'''
        usage = "status, shutter <start|stop>, mode <video|camera>, power <on|off>"
        mav = self.master.mav

        if args[0] == "status":
            self.cmd_gopro_status(args[1:])
            return

        if args[0] == "shutter":
            name = args[1].lower()
            if name == 'start':
                mav.gopro_set_request_send(self.target_system, mavutil.mavlink.MAV_COMP_ID_GIMBAL,
                 mavutil.mavlink.GOPRO_COMMAND_SHUTTER, [1, 0 ,0 , 0])
                return
            elif name == 'stop':
                mav.gopro_set_request_send(self.target_system, mavutil.mavlink.MAV_COMP_ID_GIMBAL,
                 mavutil.mavlink.GOPRO_COMMAND_SHUTTER, [0, 0 ,0 , 0])
                return
            else:
                print("unrecognized")
                return

        if args[0] == "mode":
            name = args[1].lower()
            if name == 'video':
                mav.gopro_set_request_send(self.target_system, mavutil.mavlink.MAV_COMP_ID_GIMBAL,
                 mavutil.mavlink.GOPRO_COMMAND_CAPTURE_MODE, [0, 0 ,0 , 0])
                return
            elif name == 'camera':
                mav.gopro_set_request_send(self.target_system, mavutil.mavlink.MAV_COMP_ID_GIMBAL,
                 mavutil.mavlink.GOPRO_COMMAND_CAPTURE_MODE, [1, 0 ,0 , 0])
                return
            else:
                print("unrecognized")
                return

        if args[0] == "power":
            name = args[1].lower()
            if name == 'on':
                mav.gopro_set_request_send(self.target_system, mavutil.mavlink.MAV_COMP_ID_GIMBAL,
                 mavutil.mavlink.GOPRO_COMMAND_POWER, [1, 0 ,0 , 0])
                return
            elif name == 'off':
                mav.gopro_set_request_send(self.target_system, mavutil.mavlink.MAV_COMP_ID_GIMBAL,
                 mavutil.mavlink.GOPRO_COMMAND_POWER, [0, 0 ,0 , 0])
                return
            else:
                print("unrecognized")
                return

        print(usage)

    def cmd_gopro_status(self, args):
        '''show gopro status'''
        master = self.master
        if 'GOPRO_HEARTBEAT' in master.messages:
            print(master.messages['GOPRO_HEARTBEAT'])
        else:
            print("No GOPRO_HEARTBEAT messages")

def init(mpstate):
    '''initialise module'''
    return GoProModule(mpstate)



#!/usr/bin/env python
'''
HIL module
Andrew Tridgell
December 2012

This interfaces to Tools/autotest/jsbsim/runsim.py to run the JSBSim flight simulator
'''

import sys, os, time, socket, errno, struct, math
from math import degrees, radians
from MAVProxy.modules.lib import mp_module
from pymavlink import mavutil

class HILModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(HILModule, self).__init__(mpstate, "HIL", "HIL simulation")
        self.last_sim_send_time = time.time()
        self.last_apm_send_time = time.time()
        self.rc_channels_scaled = mavutil.mavlink.MAVLink_rc_channels_scaled_message(0, 0, 0, 0, -10000, 0, 0, 0, 0, 0, 0)
        self.hil_state_msg = None
        sim_in_address  = ('127.0.0.1', 5501)
        sim_out_address  = ('127.0.0.1', 5502)

        self.sim_in = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sim_in.bind(sim_in_address)
        self.sim_in.setblocking(0)

        self.sim_out = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sim_out.connect(sim_out_address)
        self.sim_out.setblocking(0)

        # HIL needs very fast idle loop calls
        if self.settings.select_timeout > 0.001:
            self.settings.select_timeout = 0.001

    def unload(self):
        '''unload module'''
        self.sim_in.close()
        self.sim_out.close()

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == 'RC_CHANNELS_SCALED':
            self.rc_channels_scaled = m

    def idle_task(self):
        '''called from main loop'''
        self.check_sim_in()
        self.check_sim_out()
        self.check_apm_out()

    def check_sim_in(self):
        '''check for FDM packets from runsim'''
        try:
            pkt = self.sim_in.recv(17*8 + 4)
        except socket.error as e:
            if not e.errno in [ errno.EAGAIN, errno.EWOULDBLOCK ]:
                raise
            return
        if len(pkt) != 17*8 + 4:
            # wrong size, discard it
            print("wrong size %u" % len(pkt))
            return
        (latitude, longitude, altitude, heading, v_north, v_east, v_down,
         ax, ay, az,
         phidot, thetadot, psidot,
         roll, pitch, yaw,
         vcas, check) = struct.unpack('<17dI', pkt)
        (p, q, r) = self.convert_body_frame(radians(roll), radians(pitch), radians(phidot), radians(thetadot), radians(psidot))

        try:
            self.hil_state_msg = self.master.mav.hil_state_encode(int(time.time()*1e6),
                                                                        radians(roll),
                                                                        radians(pitch),
                                                                        radians(yaw),
                                                                        p,
                                                                        q,
                                                                        r,
                                                                        int(latitude*1.0e7),
                                                                        int(longitude*1.0e7),
                                                                        int(altitude*1.0e3),
                                                                        int(v_north*100),
                                                                        int(v_east*100),
                                                                        0,
                                                                        int(ax*1000/9.81),
                                                                        int(ay*1000/9.81),
                                                                        int(az*1000/9.81))
        except Exception:
            return




    def check_sim_out(self):
        '''check if we should send new servos to flightgear'''
        now = time.time()
        if now - self.last_sim_send_time < 0.02 or self.rc_channels_scaled is None:
            return
        self.last_sim_send_time = now

        servos = []
        for ch in range(1,9):
            servos.append(self.scale_channel(ch, getattr(self.rc_channels_scaled, 'chan%u_scaled' % ch)))
        servos.extend([0,0,0, 0,0,0])
        buf = struct.pack('<14H', *servos)
        try:
            self.sim_out.send(buf)
        except socket.error as e:
            if not e.errno in [ errno.ECONNREFUSED ]:
                raise
            return


    def check_apm_out(self):
        '''check if we should send new data to the APM'''
        now = time.time()
        if now - self.last_apm_send_time < 0.02:
            return
        self.last_apm_send_time = now
        if self.hil_state_msg is not None:
            self.master.mav.send(self.hil_state_msg)

    def convert_body_frame(self, phi, theta, phiDot, thetaDot, psiDot):
        '''convert a set of roll rates from earth frame to body frame'''
        p = phiDot - psiDot*math.sin(theta)
        q = math.cos(phi)*thetaDot + math.sin(phi)*psiDot*math.cos(theta)
        r = math.cos(phi)*psiDot*math.cos(theta) - math.sin(phi)*thetaDot
        return (p, q, r)

    def scale_channel(self, ch, value):
        '''scale a channel to 1000/1500/2000'''
        v = value/10000.0
        if v < -1:
            v = -1
        elif v > 1:
            v = 1
        if ch == 3 and self.mpstate.vehicle_type != 'rover':
            if v < 0:
                v = 0
            return int(1000 + v*1000)
        return int(1500 + v*500)

def init(mpstate):
    '''initialise module'''
    return HILModule(mpstate)



#!/usr/bin/env python
'''
Module to read DTM files published by Geoscience Australia
Written by Stephen Dade (stephen_dade@hotmail.com
'''

import os
import sys

import numpy


class ERMap:
    '''Class to read GA files'''
    def __init__(self):
        self.header = None
        self.data = None
        self.startlongitude = 0
        self.startlatitude = 0
        self.endlongitude = 0
        self.endlatitude = 0
        self.deltalongitude = 0
        self.deltalatitude = 0

    def read_ermapper(self, ifile):
        '''Read in a DEM file and associated .ers file'''
        ers_index = ifile.find('.ers')
        if ers_index > 0:
            data_file = ifile[0:ers_index]
            header_file = ifile
        else:
            data_file = ifile
            header_file = ifile + '.ers'

        self.header = self.read_ermapper_header(header_file)

        nroflines = int(self.header['nroflines'])
        nrofcellsperlines = int(self.header['nrofcellsperline'])
        self.data = self.read_ermapper_data(data_file, offset=int(self.header['headeroffset']))
        self.data = numpy.reshape(self.data,(nroflines,nrofcellsperlines))

        longy =  numpy.fromstring(self.getHeaderParam('longitude'), sep=':')
        latty =  numpy.fromstring(self.getHeaderParam('latitude'), sep=':')

        self.deltalatitude = float(self.header['ydimension'])
        self.deltalongitude = float(self.header['xdimension'])

        if longy[0] < 0:
            self.startlongitude = longy[0]+-((longy[1]/60)+(longy[2]/3600))
            self.endlongitude = self.startlongitude - int(self.header['nrofcellsperline'])*self.deltalongitude
        else:
            self.startlongitude = longy[0]+(longy[1]/60)+(longy[2]/3600)
            self.endlongitude = self.startlongitude + int(self.header['nrofcellsperline'])*self.deltalongitude
        if latty[0] < 0:
            self.startlatitude = latty[0]-((latty[1]/60)+(latty[2]/3600))
            self.endlatitude = self.startlatitude - int(self.header['nroflines'])*self.deltalatitude
        else:
            self.startlatitude = latty[0]+(latty[1]/60)+(latty[2]/3600)
            self.endlatitude = self.startlatitude + int(self.header['nroflines'])*self.deltalatitude

    def read_ermapper_header(self, ifile):
        # function for reading an ERMapper header from file
        header = {}

        fid = open(ifile,'rt')
        header_string = fid.readlines()
        fid.close()

        for line in header_string:
            if line.find('=') > 0:
                tmp_string = line.strip().split('=')
                header[tmp_string[0].strip().lower()]= tmp_string[1].strip()

        return header

    def read_ermapper_data(self, ifile, data_format = numpy.float32, offset=0):
        # open input file in a binary format and read the input string
        fid = open(ifile,'rb')
        if offset != 0:
            fid.seek(offset)
        input_string = fid.read()
        fid.close()

        # convert input string to required format (Note default format is numpy.float32)
        grid_as_float = numpy.fromstring(input_string,data_format)
        return grid_as_float

    def getHeaderParam(self, key):
         '''Find a parameter in the associated .ers file'''
         return self.header[key]

    def printBoundingBox(self):
        '''Print the bounding box that this DEM covers'''
        print("Bounding Latitude: ")
        print(self.startlatitude)
        print(self.endlatitude)

        print("Bounding Longitude: ")
        print(self.startlongitude)
        print(self.endlongitude)

    def getPercentBlank(self):
        '''Print how many null cells are in the DEM - Quality measure'''
        blank = 0
        nonblank = 0
        for x in self.data.flat:
            if x == -99999.0:
                blank = blank + 1
            else:
                nonblank = nonblank + 1

        print("Blank tiles =  ", blank, "out of ", (nonblank+blank))

    def getAltitudeAtPoint(self, latty, longy):
        '''Return the altitude at a particular long/lat'''
        #check the bounds
        if self.startlongitude > 0 and (longy < self.startlongitude or longy > self.endlongitude):
            return -1
        if self.startlongitude < 0 and (longy > self.startlongitude or longy < self.endlongitude):
            return -1
        if self.startlatitude > 0 and (latty < self.startlatitude or longy > self.endlatitude):
            return -1
        if self.startlatitude < 0 and (latty > self.startlatitude or longy < self.endlatitude):
            return -1

        x = numpy.abs((latty - self.startlatitude)/self.deltalatitude)
        y = numpy.abs((longy - self.startlongitude)/self.deltalongitude)

        #do some interpolation
        # print("x,y", x, y)
        x_int = int(x)
        x_frac = x - int(x)
        y_int = int(y)
        y_frac = y - int(y)
        #print("frac", x_int, x_frac, y_int, y_frac)
        value00 = self.data[x_int, y_int]
        value10 = self.data[x_int+1, y_int]
        value01 = self.data[x_int, y_int+1]
        value11 = self.data[x_int+1, y_int+1]
        #print("values ", value00, value10, value01, value11)

        #check for null values
        if value00 == -99999:
            value00 = 0
        if value10 == -99999:
            value10 = 0
        if value01 == -99999:
            value01 = 0
        if value11 == -99999:
            value11 = 0

        value1 = self._avg(value00, value10, x_frac)
        value2 = self._avg(value01, value11, x_frac)
        value  = self._avg(value1,  value2, y_frac)

        return value

    @staticmethod
    def _avg(value1, value2, weight):
        """Returns the weighted average of two values and handles the case where
            one value is None. If both values are None, None is returned.
        """
        if value1 is None:
            return value2
        if value2 is None:
            return value1
        return value2 * weight + value1 * (1 - weight)



if __name__ == '__main__':

    print("./Canberra/GSNSW_P756demg")
    mappy = ERMap()
    mappy.read_ermapper(os.path.join(os.environ['HOME'], './Documents/Elevation/Canberra/GSNSW_P756demg'))

    #print some header data
    mappy.printBoundingBox()

    #get a measure of data quality
    #mappy.getPercentBlank()

    #test the altitude (around Canberra):
    alt = mappy.getAltitudeAtPoint(-35.274411, 149.097504)
    print("Alt at (-35.274411, 149.097504) is 807m (Google) or " + str(alt))
    alt = mappy.getAltitudeAtPoint(-35.239648, 149.126118)
    print("Alt at (-35.239648, 149.126118) is 577m (Google) or " + str(alt))
    alt = mappy.getAltitudeAtPoint(-35.362751, 149.165361)
    print("Alt at (-35.362751, 149.165361) is 584m (Google) or " + str(alt))
    alt = mappy.getAltitudeAtPoint(-35.306992, 149.194274)
    print("Alt at (-35.306992, 149.194274) is 570m (Google) or " + str(alt))
    alt = mappy.getAltitudeAtPoint(-35.261612, 149.542091)
    print("Alt at (-35.261612, 149.542091) is 766m (Google) or " + str(alt))
    alt = mappy.getAltitudeAtPoint(-35.052544, 149.509165)
    print("Alt at (-35.052544, 149.509165) is 700m (Google) or " + str(alt))
    alt = mappy.getAltitudeAtPoint(-35.045126, 149.257482)
    print("Alt at (-35.045126, 149.257482) is 577m (Google) or " + str(alt))
    alt = mappy.getAltitudeAtPoint(-35.564044, 149.177657)
    print("Alt at (-35.564044, 149.177657) is 1113m (Google) or " + str(alt))







# Maintain the guard for wx being able to be
# loaded on this process. This is only needed
# for Darwin, so to reduce error cases, only
# implement it on Darwin.

import sys
safe = sys.platform != 'darwin'



#!/usr/bin/env python
'''
Antenna tracker control module
This module catches MAVLINK_MSG_ID_GLOBAL_POSITION_INT
and sends them to a MAVlink connected antenna tracker running
ardupilot AntennaTracker
Mike McCauley, based on earlier work by Andrew Tridgell
June 2012
'''

import sys, os, time
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules import mavproxy_map
from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.mavproxy_param import ParamState

# this should be in mavutil.py
mode_mapping_antenna = {
    'MANUAL' : 0,
    'AUTO' : 10,
    'INITIALISING' : 16
    }

class TrackerModule(mp_module.MPModule):
    def __init__(self, mpstate):
        from pymavlink import mavparm
        super(TrackerModule, self).__init__(mpstate, "tracker", "antenna tracker control module")
        self.connection = None
        self.tracker_param = mavparm.MAVParmDict()
        sysid = 2
        self.pstate = ParamState(self.tracker_param, self.logdir, self.vehicle_name, 'tracker.parm', mpstate, sysid)
        self.tracker_settings = mp_settings.MPSettings(
            [ ('port', str, "/dev/ttyUSB0"),
              ('baudrate', int, 57600),
              ('debug', int, 0)
              ]
            )
        self.add_command('tracker', self.cmd_tracker,
                         "antenna tracker control module",
                         ['<start|arm|disarm|level|mode|position|calpress|mode>',
                          'set (TRACKERSETTING)',
                          'param <set|show|fetch|help> (TRACKERPARAMETER)',
                          'param (TRACKERSETTING)'])
        self.add_completion_function('(TRACKERSETTING)', self.tracker_settings.completion)
        self.add_completion_function('(TRACKERPARAMETER)', self.complete_parameter)

    def complete_parameter(self, text):
        '''complete a tracker parameter'''
        return self.tracker_param.keys()

    def find_connection(self):
        '''find an antenna tracker connection if possible'''
        if self.connection is not None:
            return self.connection
        for m in self.mpstate.mav_master:
            if 'HEARTBEAT' in m.messages:
                if m.messages['HEARTBEAT'].type == mavutil.mavlink.MAV_TYPE_ANTENNA_TRACKER:
                    return m
        return None

    def cmd_tracker(self, args):
        '''tracker command parser'''
        usage = "usage: tracker <start|set|arm|disarm|level|param|mode|position> [options]"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "start":
            self.cmd_tracker_start()
        elif args[0] == "set":
            self.tracker_settings.command(args[1:])
        elif args[0] == 'arm':
            self.cmd_tracker_arm()
        elif args[0] == 'disarm':
            self.cmd_tracker_disarm()
        elif args[0] == 'level':
            self.cmd_tracker_level()
        elif args[0] == 'param':
            self.cmd_tracker_param(args[1:])
        elif args[0] == 'mode':
            self.cmd_tracker_mode(args[1:])
        elif args[0] == 'position':
            self.cmd_tracker_position(args[1:])
        elif args[0] == 'calpress':
            self.cmd_tracker_calpress(args[1:])
        else:
            print(usage)

    def cmd_tracker_position(self, args):
        '''tracker manual positioning commands'''
        connection = self.find_connection()
        if not connection:
            print("No antenna tracker found")
            return
        positions = [0, 0, 0, 0, 0] # x, y, z, r, buttons. only position[0] (yaw) and position[1] (pitch) are currently used
        for i in range(0, 4):
            if len(args) > i:
                positions[i] = int(args[i]) # default values are 0
        connection.mav.manual_control_send(connection.target_system,
                                           positions[0], positions[1],
                                           positions[2], positions[3],
                                           positions[4])

    def cmd_tracker_calpress(self, args):
        '''calibrate barometer on tracker'''
        connection = self.find_connection()
        if not connection:
            print("No antenna tracker found")
            return
        connection.calibrate_pressure()

    def cmd_tracker_mode(self, args):
        '''set arbitrary mode'''
        connection = self.find_connection()
        if not connection:
            print("No antenna tracker found")
            return
        mode_mapping = connection.mode_mapping()
        if mode_mapping is None:
            print('No mode mapping available')
            return
        if len(args) != 1:
            print('Available modes: ', mode_mapping.keys())
            return
        mode = args[0].upper()
        if mode not in mode_mapping:
            print('Unknown mode %s: ' % mode)
            return
        connection.set_mode(mode_mapping[mode])

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet from the master vehicle. Relay it to the tracker
        if it is a GLOBAL_POSITION_INT'''
        if m.get_type() in ['GLOBAL_POSITION_INT', 'SCALED_PRESSURE']:
            connection = self.find_connection()
            if not connection:
                return
            if m.get_srcSystem() != connection.target_system:
                connection.mav.send(m)

    def idle_task(self):
        '''called in idle time'''
        if not self.connection:
            return

        # check for a mavlink message from the tracker
        m = self.connection.recv_msg()
        if m is None:
            return

        if self.tracker_settings.debug:
            print(m)

        self.pstate.handle_mavlink_packet(self.connection, m)
        self.pstate.fetch_check(self.connection)

        if self.module('map') is None:
            return

        if m.get_type() == 'GLOBAL_POSITION_INT':
            (self.lat, self.lon, self.heading) = (m.lat*1.0e-7, m.lon*1.0e-7, m.hdg*0.01)
            if self.lat != 0 or self.lon != 0:
                self.module('map').create_vehicle_icon('AntennaTracker', 'red', follow=False, vehicle_type='antenna')
                self.mpstate.map.set_position('AntennaTracker', (self.lat, self.lon), rotation=self.heading)


    def cmd_tracker_start(self):
        if self.tracker_settings.port is None:
            print("tracker port not set")
            return
        if self.connection is not None:
            self.connection.close()
            self.connection = None
            print("Closed old connection")
        print("connecting to tracker %s at %d" % (self.tracker_settings.port,
                                                  self.tracker_settings.baudrate))
        m = mavutil.mavlink_connection(self.tracker_settings.port,
                                       autoreconnect=True,
                                       source_system=self.settings.source_system,
                                       baud=self.tracker_settings.baudrate)
        m.mav.srcComponent = self.settings.source_component
        if self.logdir:
            m.setup_logfile(os.path.join(self.logdir, 'tracker.tlog'))
        self.connection = m

    def cmd_tracker_arm(self):
        '''Enable the servos in the tracker so the antenna will move'''
        if not self.connection:
            print("tracker not connected")
            return
        self.connection.arducopter_arm()

    def cmd_tracker_disarm(self):
        '''Disable the servos in the tracker so the antenna will not move'''
        if not self.connection:
            print("tracker not connected")
            return
        self.connection.arducopter_disarm()

    def cmd_tracker_level(self):
        '''Calibrate the accelerometers. Disarm and move the antenna level first'''
        if not self.connection:
            print("tracker not connected")
            return
        self.connection.calibrate_level()

    def cmd_tracker_param(self, args):
        '''Parameter commands'''
        if not self.connection:
            print("tracker not connected")
            return
        self.pstate.handle_command(self.connection, self.mpstate, args)

def init(mpstate):
    '''initialise module'''
    return TrackerModule(mpstate)



#!/usr/bin/env python
'''
Checklist module
Stephen Dade
July 2014
'''

import math, sys, os, time
from MAVProxy.modules.lib import mp_checklist
from MAVProxy.modules.lib import mp_module
from pymavlink import mavutil

class ChecklistModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(ChecklistModule, self).__init__(mpstate, "checklist", "checklist module")
        self.checklist = mp_checklist.CheckUI()

    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        if not isinstance(self.checklist, mp_checklist.CheckUI):
            return
        if not self.checklist.is_alive():
            return

        type = msg.get_type()
        master = self.master

        if type == 'HEARTBEAT':
            '''beforeEngineList - APM booted'''
            if self.mpstate.status.heartbeat_error == True:
                self.checklist.set_check("Pixhawk Booted", 0)
            else:
                self.checklist.set_check("Pixhawk Booted", 1)

        '''beforeEngineList - Flight mode MANUAL'''
        if self.mpstate.status.flightmode == "MANUAL":
            self.checklist.set_check("Flight mode MANUAL", 1)
        else:
            self.checklist.set_check("Flight mode MANUAL", 0)

        if type in [ 'GPS_RAW', 'GPS_RAW_INT' ]:
            '''beforeEngineList - GPS lock'''
            if ((msg.fix_type >= 3 and master.mavlink10()) or
                (msg.fix_type == 2 and not master.mavlink10())):
                self.checklist.set_check("GPS lock", 1)
            else:
                self.checklist.set_check("GPS lock", 0)

        '''beforeEngineList - Radio Links > 6db margin TODO: figure out how to read db levels'''
        if type in ['RADIO', 'RADIO_STATUS']:
            if msg.rssi < msg.noise+6 or msg.remrssi < msg.remnoise+6:
                self.checklist.set_check("Radio links > 6db margin", 0)
            else:
                self.checklist.set_check("Radio Links > 6db margin", 0)

        if type == 'HWSTATUS':
            '''beforeEngineList - Avionics Battery'''
            if msg.Vcc >= 4600 and msg.Vcc <= 5300:
                self.checklist.set_check("Avionics Power", 1)
            else:
                self.checklist.set_check("Avionics Power", 0)

        if type == 'POWER_STATUS':
            '''beforeEngineList - Servo Power'''
            if msg.Vservo >= 4900 and msg.Vservo <= 6500:
                self.checklist.set_check("Servo Power", 1)
            else:
                self.checklist.set_check("Servo Power", 0)

        '''beforeEngineList - Waypoints Loaded'''
        if type == 'HEARTBEAT':
            if self.module('wp').wploader.count() == 0:
                self.checklist.set_check("Waypoints Loaded", 0)
            else:
                self.checklist.set_check("Waypoints Loaded", 1)

        '''beforeTakeoffList - Compass active'''
        if type == 'GPS_RAW':
            if math.fabs(msg.hdg - master.field('VFR_HUD', 'heading', '-')) < 10 or math.fabs(msg.hdg - master.field('VFR_HUD', 'heading', '-')) > 355:
                self.checklist.set_check("Compass active", 1)
            else:
                self.checklist.set_check("Compass active", 0)

        '''beforeCruiseList - Airspeed > 10 m/s , Altitude > 30 m'''
        if type == 'VFR_HUD':
            rel_alt = master.field('GLOBAL_POSITION_INT', 'relative_alt', 0) * 1.0e-3
            if rel_alt > 30:
                self.checklist.set_check("Altitude > 30 m", 1)
            else:
                self.checklist.set_check("Altitude > 30 m", 0)
            if msg.airspeed > 10 or msg.groundspeed > 10:
                self.checklist.set_check("Airspeed > 10 m/s", 1)
            else:
                self.checklist.set_check("Airspeed > 10 m/s", 0)

        '''beforeEngineList - IMU'''
        if type in ['SYS_STATUS']:
            sensors = { 'AS'  : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_DIFFERENTIAL_PRESSURE,
                        'MAG' : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_3D_MAG,
                        'INS' : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_3D_ACCEL | mavutil.mavlink.MAV_SYS_STATUS_SENSOR_3D_GYRO,
                        'AHRS' : mavutil.mavlink.MAV_SYS_STATUS_AHRS}
            bits = sensors['INS']
            present = ((msg.onboard_control_sensors_enabled & bits) == bits)
            healthy = ((msg.onboard_control_sensors_health & bits) == bits)
            if not present or not healthy:
                self.checklist.set_check("IMU Check", 1)
            else:
                self.checklist.set_check("IMU Check", 0)

    def unload(self):
        '''unload module'''
        self.checklist.close()
        
def init(mpstate):
    '''initialise module'''
    return ChecklistModule(mpstate)



#!/usr/bin/env python
'''
Event class and enums for Parameter Editor

'''
# ParamEditorEvents come FROM the GUI (with a few exceptions where the
# Parameter Editor Module sends a message to itself).
# ParamEditorEditorGUIEvents go to the GUI
# enum for ParamEditorEvent types
PEE_READ_KEY = 0
PEE_LOAD_FILE = 1
PEE_SAVE_FILE = 2
PEE_READ_PARAM = 3
PEE_WRITE_PARAM = 4
PEE_TIME_TO_QUIT = 5
PEE_RESET = 6
PEE_FETCH = 7
# enum of ParamEditorGUIEvent types
PEGE_READ_PARAM = 0
PEGE_REFRESH_PARAM = 1
PEGE_WRITE_SUCC = 2
PEGE_RCIN = 3


class ParamEditorEvent:
    def __init__(self, type, **kwargs):
        self.type = type
        self.arg_dict = kwargs

        if self.type not in [PEE_SAVE_FILE, PEE_LOAD_FILE, PEE_READ_KEY,
                             PEGE_REFRESH_PARAM, PEGE_READ_PARAM, PEGE_RCIN,
                             PEE_READ_PARAM, PEE_WRITE_PARAM, PEE_RESET,
                             PEE_TIME_TO_QUIT, PEGE_WRITE_SUCC, PEE_FETCH]:
            raise TypeError("Unrecongized ParamEditorEvent type:" +
                            str(self.type))

    def get_type(self):
        return self.type

    def get_arg(self, key):
        if key not in self.arg_dict:
            print("No key %s in %s" % (key, str(self.type)))
            return None
        return self.arg_dict[key]



# use optitrack data to provide ATT_POS_MOCAP data
# it works with optitrack motion capture cameras and optitrack motive tracker software (https://optitrack.com/software/motive/)
# yuan-chu tai

import time
from pymavlink import mavutil
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.mavproxy_optitrack import NatNetClient

class optitrack(mp_module.MPModule):
    def __init__(self, mpstate):
        """Initialise module"""
        super(optitrack, self).__init__(mpstate, "optitrack", "optitrack")
        self.optitrack_settings = mp_settings.MPSettings(
            [('server', str, '127.0.0.1'),
            ('client', str, '127.0.0.1'),
            ('msg_intvl_ms', int, 75),
            ('obj_id', int, 1)]
        )
        self.add_command('optitrack', self.cmd_optitrack, "optitrack control", ['<start>', 'set (OPTITRACKSETTING)'])
        self.streaming_client = NatNetClient.NatNetClient()
        # Configure the streaming client to call our rigid body handler on the emulator to send data out.
        self.streaming_client.rigid_body_listener = self.receive_rigid_body_frame
        self.last_msg_time = 0
        self.started = False

    # This is a callback function that gets connected to the NatNet client. It is called once per rigid body per frame
    def receive_rigid_body_frame(self, new_id, position, rotation, tracking_valid):
        if (tracking_valid and new_id == self.optitrack_settings.obj_id):
            now = time.time()
            if (now - self.last_msg_time) > (self.optitrack_settings.msg_intvl_ms * 0.001):
                time_us = int(now * 1.0e6)
                self.master.mav.att_pos_mocap_send(time_us, (rotation[3], rotation[0], rotation[2], -rotation[1]), position[0], position[2], -position[1])
                self.last_msg_time = now

    def usage(self):
        '''show help on command line options'''
        return "Usage: optitrack <start|set>"

    def cmd_start(self):
        self.streaming_client.set_client_address(self.optitrack_settings.client)
        self.streaming_client.set_server_address(self.optitrack_settings.server)
        self.streaming_client.setup_sdk()
        self.started = True

    def cmd_optitrack(self, args):
        '''control behaviour of the module'''
        if len(args) == 0:
            print(self.usage())
        elif args[0] == "start":
            self.cmd_start()
        elif args[0] == "set":
            self.optitrack_settings.command(args[1:])
        else:
            print(self.usage())

    def idle_task(self):
        '''called rapidly by mavproxy'''
        if self.started:
            self.streaming_client.process_data_and_cmd()

def init(mpstate):
    '''initialise module'''
    return optitrack(mpstate)



#!/usr/bin/env python
'''
param editor module
Akshath Singhal
June 2019
'''

from MAVProxy.modules.lib import mp_module


class ParamEditorModule(mp_module.MPModule):
    '''
    A Graphical parameter editor for use with MAVProxy
    '''
    def __init__(self, mpstate):
        super(ParamEditorModule, self).__init__(mpstate,
                                                "paramedit", "param edit",
                                                public=True)

        # to work around an issue on MacOS this module is a thin wrapper
        # around a separate ParamEditorMain object
        self.pe_main = None
        self.mpstate = mpstate

    def unload(self):
        '''unload module'''
        if self.pe_main:
            self.pe_main.unload()

    def idle_task(self):
        if not self.pe_main:
            # wait for parameter module to load
            if self.module('param') is None:
                return
            from MAVProxy.modules.mavproxy_paramedit import param_editor
            self.pe_main = param_editor.ParamEditorMain(self.mpstate)
        if self.pe_main:
            if self.pe_main.needs_unloading:
                self.needs_unloading = True
        self.pe_main.idle_task()

    def mavlink_packet(self, m):
        if self.pe_main:
            self.pe_main.mavlink_packet(m)


def init(mpstate):
    '''initialise module'''
    return ParamEditorModule(mpstate)



from setuptools import setup
import os, platform, sys

version = "1.8.52"

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

package_data = ['modules/mavproxy_map/data/*.jpg', 
                'modules/mavproxy_map/data/*.png',
                'modules/mavproxy_mmap/mmap_app/*',
                'modules/mavproxy_joystick/joysticks/*.yml',
                'modules/mavproxy_magical/data/*.mtl',
                'modules/mavproxy_magical/data/*.obj',
                'modules/mavproxy_fieldcheck/*.txt',
                'tools/graphs/*.xml',
]

package_data.extend(package_files('MAVProxy/modules/mavproxy_cesium/app'))

# note that we do not include all the real dependencies here (like matplotlib etc)
# as that breaks the pip install. It seems that pip is not smart enough to
# use the system versions of these dependencies, so it tries to download and install
# large numbers of modules like numpy etc which may be already installed
requirements=['pymavlink>=2.4.14',
              'pyserial>=3.0']

if platform.system() == "Darwin":
    # on MacOS we can have a more complete requirements list
    requirements.extend(['billiard>=3.5.0',
                         'gnureadline',
                         'matplotlib',
                         'numpy',
                         'opencv-python',
                         'lxml',
                         'future',
                         'wxPython'])

if platform.system() == "Windows" and sys.version_info >= (3, 0):
    # on MacOS we can have a more complete requirements list
    requirements.extend(['prompt_toolkit'])
elif platform.system() == "Windows":
    requirements.extend(['pyreadline'])

setup(name='MAVProxy',
      version=version,
      zip_safe=True,
      description='MAVProxy MAVLink ground station',
      long_description='''A MAVLink protocol proxy and ground station. MAVProxy
is oriented towards command line operation, and is suitable for embedding in
small autonomous vehicles or for using on ground control stations. It also
features a number of graphical tools such as a slipmap for satellite mapping
view of the vehicles location, and status console and several useful vehicle
control modules. MAVProxy is extensible via a modules system - see the modules
subdirectory for some example modules. MAVProxy was developed by CanberraUAV
for use in the 2012 Outback Challenge, and includes a module for the
CanberraUAV search and rescue system. See
https://ardupilot.org/mavproxy/index.html for more information
on how to use MAVProxy.''',
      url='https://github.com/ArduPilot/MAVProxy',
      author='Andrew Tridgell',
      author_email='andrew@tridgell.net',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering'],
      license='GPLv3',
      packages=['MAVProxy',
                'MAVProxy.modules',
                'MAVProxy.modules.mavproxy_fieldcheck',
                'MAVProxy.modules.mavproxy_map',
                'MAVProxy.modules.mavproxy_mmap',
                'MAVProxy.modules.mavproxy_misseditor',
                'MAVProxy.modules.mavproxy_paramedit',
                'MAVProxy.modules.mavproxy_smartcamera',
                'MAVProxy.modules.mavproxy_cesium',
                'MAVProxy.modules.mavproxy_joystick',
                'MAVProxy.modules.mavproxy_magical',
                'MAVProxy.modules.mavproxy_optitrack',
                'MAVProxy.modules.mavproxy_nokov',
                'MAVProxy.modules.lib',
                'MAVProxy.modules.lib.ANUGA',
                'MAVProxy.modules.lib.MacOS',
                'MAVProxy.modules.lib.optparse_gui'],
      install_requires=requirements,
      extras_require={
        # restserver module
        'server': ['flask'],
      },
      scripts=['MAVProxy/mavproxy.py',
               'MAVProxy/tools/mavflightview.py',
               'MAVProxy/tools/MAVExplorer.py',
               'MAVProxy/modules/mavproxy_map/mp_slipmap.py',
               'MAVProxy/modules/mavproxy_map/mp_tile.py'],
      package_data={'MAVProxy':
                    package_data}
    )



#!/usr/bin/env python
'''
mission editor module
Michael Day
June 2104
'''

from MAVProxy.modules.lib import mp_module

class MissionEditorModule(mp_module.MPModule):
    '''
    A Mission Editor for use with MAVProxy
    '''
    def __init__(self, mpstate):
        super(MissionEditorModule, self).__init__(mpstate, "misseditor", "mission editor", public = True)

        # to work around an issue on MacOS this module is a thin wrapper around a separate MissionEditorMain object
        from MAVProxy.modules.mavproxy_misseditor import mission_editor
        self.me_main = mission_editor.MissionEditorMain(mpstate)

    def unload(self):
        '''unload module'''
        self.me_main.unload()

    def idle_task(self):
        self.me_main.idle_task()
        if self.me_main.needs_unloading:
            self.needs_unloading = True

    def mavlink_packet(self, m):
        self.me_main.mavlink_packet(m)

    def click_updated(self):
        self.me_main.update_map_click_position(self.mpstate.click_location)

def init(mpstate):
    '''initialise module'''
    return MissionEditorModule(mpstate)



#!/usr/bin/env python
'''RC min/max setup'''

from MAVProxy.modules.lib import mp_module

class RCSetupModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(RCSetupModule, self).__init__(mpstate, "rcsetup")
        self.calibrating = False
        self.num_channels = 4
        self.clear_rc_cal()
        self.add_command('rccal', self.cmd_rccal, "RC calibration start/stop")
        self.add_command('rctrim', self.cmd_rctrim, "RC min/max trim")
        self.empty_input_count = None
        print("rcsetup initialised")

    def clear_rc_cal(self):
        self.rc_cal = []
        self.rc_cal.append("") # 0 will be empty
        for i in range(1,self.num_channels+1):
            #min, max, modified
            self.rc_cal.append([1500, 1500, False])

    def apply_rc_cal(self):
        for i in range(1, len(self.rc_cal)):
            #only apply calibration changes to channels that
            #were modified during calibration
            if (self.rc_cal[i][2] == False):
                continue

            self.param_set('RC%u_MIN' % i, self.rc_cal[i][0], 5)
            self.console.writeln("Set: RC%u_MIN=%u" % (i, self.rc_cal[i][0]))
            self.param_set('RC%u_MAX' % i, self.rc_cal[i][1], 5)
            self.console.writeln("Set: RC%u_MAX=%u" % (i, self.rc_cal[i][1]))

    def get_cal_min(self, channel):
        return self.rc_cal[channel][0]

    def get_cal_max(self, channel):
        return self.rc_cal[channel][1]

    def set_cal_min(self, channel, val):
        self.rc_cal[channel][0] = val
        self.rc_cal[channel][2] = True

    def set_cal_max(self, channel, val):
        self.rc_cal[channel][1] = val
        self.rc_cal[channel][2] = True

    def cmd_rccal(self, args):
        '''start/stop RC calibration'''
        if len(args) < 1:
            self.print_cal_usage()
            return

        if (args[0] == "start"):
            if len(args) > 1:
                self.num_channels = int(args[1])
            print("Calibrating %u channels" % self.num_channels)
            print("WARNING: remove propellers from electric planes!!")
            print("Push return when ready to calibrate.")
            self.empty_input_count = self.mpstate.empty_input_count
        elif (args[0] == "done"):
            self.calibrating = False
            self.apply_rc_cal()
        else:
            self.print_cal_usage()

    def idle_task(self):
        if self.empty_input_count is None:
            # not waiting for input
            return
        if self.empty_input_count == self.mpstate.empty_input_count:
            # haven't received input
            return

        self.clear_rc_cal()
        self.calibrating = True
        self.empty_input_count = None

    def cmd_rctrim(self, args):
        '''set RCx_TRIM'''
        if not 'RC_CHANNELS' in self.status.msgs:
            print("No RC_CHANNELS to trim with")
            return
        m = self.status.msgs['RC_CHANNELS']
        for ch in range(1,5):
            self.param_set('RC%u_TRIM' % ch, getattr(m, 'chan%u_raw' % ch))


    def unload(self):
        if 'rcreset' in self.mpstate.command_map:
            self.mpstate.command_map.pop('rcreset')
        if 'rctrim' in self.mpstate.command_map:
            self.mpstate.command_map.pop('rctrim')


    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        #do nothing if not caibrating
        if (self.calibrating == False):
            return

        if m.get_type() == 'RC_CHANNELS_RAW':
            for i in range(1,self.num_channels+1):
                v = getattr(m, 'chan%u_raw' % i)

                if self.get_cal_min(i) > v:
                    self.set_cal_min(i,v)
                    self.console.writeln("Calibrating: RC%u_MIN=%u" % (i, v))
                if self.get_cal_max(i) < v:
                    self.set_cal_max(i,v)
                    self.console.writeln("Calibrating: RC%u_MAX=%u" % (i, v))

    def print_cal_usage(self):
        print("Usage rccal <start|done>")

def init(mpstate):
    '''initialise module'''
    return RCSetupModule(mpstate)
